from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence
from uuid import UUID

import cloudpickle  # type: ignore[import-untyped]

import pixeltable.exceptions as excs
from pixeltable.utils.coroutine import run_coroutine_synchronously

from .function import Function
from .signature import Signature

if TYPE_CHECKING:
    from pixeltable import exprs


class CallableFunction(Function):
    """Pixeltable Function backed by a Python Callable.

    CallableFunctions come in two flavors:
    - references to lambdas and functions defined in notebooks, which are pickled and serialized to the store
    - functions that are defined in modules are serialized via the default mechanism
    """

    py_fns: list[Callable]
    self_name: Optional[str]
    batch_size: Optional[int]

    def __init__(
        self,
        signatures: list[Signature],
        py_fns: list[Callable],
        self_path: Optional[str] = None,
        self_name: Optional[str] = None,
        batch_size: Optional[int] = None,
        is_method: bool = False,
        is_property: bool = False,
    ):
        assert len(signatures) > 0
        assert len(signatures) == len(py_fns)
        if self_path is None and len(signatures) > 1:
            raise excs.Error('Multiple signatures are only allowed for module UDFs (not locally defined UDFs)')
        self.py_fns = py_fns
        self.self_name = self_name
        self.batch_size = batch_size
        self.__doc__ = self.py_fns[0].__doc__
        super().__init__(signatures, self_path=self_path, is_method=is_method, is_property=is_property)

    def _update_as_overload_resolution(self, signature_idx: int) -> None:
        assert len(self.py_fns) > signature_idx
        self.py_fns = [self.py_fns[signature_idx]]

    @property
    def is_batched(self) -> bool:
        return self.batch_size is not None

    @property
    def is_async(self) -> bool:
        return inspect.iscoroutinefunction(self.py_fn)

    def comment(self) -> Optional[str]:
        return inspect.getdoc(self.py_fns[0])

    @property
    def py_fn(self) -> Callable:
        assert not self.is_polymorphic
        return self.py_fns[0]

    async def aexec(self, *args: Any, **kwargs: Any) -> Any:
        assert not self.is_polymorphic
        assert self.is_async
        if self.is_batched:
            # Pack the batched parameters into singleton lists
            constant_param_names = [p.name for p in self.signature.constant_parameters]
            batched_args = [[arg] for arg in args]
            constant_kwargs = {k: v for k, v in kwargs.items() if k in constant_param_names}
            batched_kwargs = {k: [v] for k, v in kwargs.items() if k not in constant_param_names}
            result = await self.py_fn(*batched_args, **constant_kwargs, **batched_kwargs)
            assert len(result) == 1
            return result[0]
        else:
            return await self.py_fn(*args, **kwargs)

    def exec(self, args: Sequence[Any], kwargs: dict[str, Any]) -> Any:
        assert not self.is_polymorphic
        if self.is_batched:
            # Pack the batched parameters into singleton lists
            constant_param_names = [p.name for p in self.signature.constant_parameters]
            batched_args = [[arg] for arg in args]
            constant_kwargs = {k: v for k, v in kwargs.items() if k in constant_param_names}
            batched_kwargs = {k: [v] for k, v in kwargs.items() if k not in constant_param_names}
            result: list[Any]
            if inspect.iscoroutinefunction(self.py_fn):
                # TODO: This is temporary (see note in utils/coroutine.py)
                result = run_coroutine_synchronously(self.py_fn(*batched_args, **constant_kwargs, **batched_kwargs))
            else:
                result = self.py_fn(*batched_args, **constant_kwargs, **batched_kwargs)
            assert len(result) == 1
            return result[0]
        elif inspect.iscoroutinefunction(self.py_fn):
            # TODO: This is temporary (see note in utils/coroutine.py)
            return run_coroutine_synchronously(self.py_fn(*args, **kwargs))
        else:
            return self.py_fn(*args, **kwargs)

    async def aexec_batch(self, *args: Any, **kwargs: Any) -> list:
        """Execute the function with the given arguments and return the result.
        The arguments are expected to be batched: if the corresponding parameter has type T,
        then the argument should have type T if it's a constant parameter, or list[T] if it's
        a batched parameter.
        """
        assert self.is_batched
        assert self.is_async
        assert not self.is_polymorphic
        # Unpack the constant parameters
        constant_kwargs, batched_kwargs = self.create_batch_kwargs(kwargs)
        return await self.py_fn(*args, **constant_kwargs, **batched_kwargs)

    def exec_batch(self, args: list[Any], kwargs: dict[str, Any]) -> list:
        """Execute the function with the given arguments and return the result.
        The arguments are expected to be batched: if the corresponding parameter has type T,
        then the argument should have type T if it's a constant parameter, or list[T] if it's
        a batched parameter.
        """
        assert self.is_batched
        assert not self.is_polymorphic
        assert not self.is_async
        # Unpack the constant parameters
        constant_kwargs, batched_kwargs = self.create_batch_kwargs(kwargs)
        return self.py_fn(*args, **constant_kwargs, **batched_kwargs)

    def create_batch_kwargs(self, kwargs: dict[str, Any]) -> tuple[dict[str, Any], dict[str, list[Any]]]:
        """Converts kwargs containing lists into constant and batched kwargs in the format expected by a batched udf."""
        constant_param_names = [p.name for p in self.signature.constant_parameters]
        constant_kwargs = {k: v[0] for k, v in kwargs.items() if k in constant_param_names}
        batched_kwargs = {k: v for k, v in kwargs.items() if k not in constant_param_names}
        return constant_kwargs, batched_kwargs

    def get_batch_size(self, *args: Any, **kwargs: Any) -> Optional[int]:
        return self.batch_size

    @property
    def display_name(self) -> str:
        return self.self_name

    @property
    def name(self) -> str:
        return self.self_name

    def overload(self, fn: Callable) -> CallableFunction:
        if self.self_path is None:
            raise excs.Error('`overload` can only be used with module UDFs (not locally defined UDFs)')
        if self.is_method or self.is_property:
            raise excs.Error('`overload` cannot be used with `is_method` or `is_property`')
        if self._has_resolved_fns:
            raise excs.Error('New `overload` not allowed after the UDF has already been called')
        if self._conditional_return_type is not None:
            raise excs.Error('New `overload` not allowed after a conditional return type has been specified')
        sig = Signature.create(fn)
        self.signatures.append(sig)
        self.py_fns.append(fn)
        return self

    def _as_dict(self) -> dict:
        if self.self_path is None:
            # this is not a module function
            assert not self.is_method and not self.is_property
            from .function_registry import FunctionRegistry

            id = FunctionRegistry.get().create_stored_function(self)
            return {'id': id.hex}
        return super()._as_dict()

    @classmethod
    def _from_dict(cls, d: dict) -> Function:
        if 'id' in d:
            from .function_registry import FunctionRegistry

            return FunctionRegistry.get().get_stored_function(UUID(hex=d['id']))
        return super()._from_dict(d)

    def to_store(self) -> tuple[dict, bytes]:
        assert not self.is_polymorphic  # multi-signature UDFs not allowed for stored fns
        md = {'signature': self.signature.as_dict(), 'batch_size': self.batch_size}
        return md, cloudpickle.dumps(self.py_fn)

    @classmethod
    def from_store(cls, name: Optional[str], md: dict, binary_obj: bytes) -> Function:
        py_fn = cloudpickle.loads(binary_obj)
        assert callable(py_fn)
        sig = Signature.from_dict(md['signature'])
        batch_size = md['batch_size']
        return CallableFunction([sig], [py_fn], self_name=name, batch_size=batch_size)

    def validate_call(self, bound_args: dict[str, 'exprs.Expr']) -> None:
        from pixeltable import exprs

        super().validate_call(bound_args)
        if self.is_batched:
            signature = self.signatures[0]
            for param in signature.constant_parameters:
                # Check that constant parameters map to constant arguments. It's ok for the argument to be a Variable,
                # since in that case the FunctionCall is part of an unresolved template; the check will be done again
                # when the template is fully resolved.
                if param.name in bound_args and not isinstance(bound_args[param.name], (exprs.Literal, exprs.Variable)):
                    raise ValueError(f'{self.display_name}(): parameter {param.name} must be a constant value')

    def __repr__(self) -> str:
        return f'<Pixeltable UDF {self.name}>'



import dataclasses
import os
from typing import Optional

import bs4
import fitz  # type: ignore[import-untyped]
import puremagic

from pixeltable import exceptions as excs, type_system as ts
from pixeltable.env import Env


@dataclasses.dataclass
class DocumentHandle:
    format: ts.DocumentType.DocumentFormat
    bs_doc: Optional[bs4.BeautifulSoup] = None
    md_ast: Optional[dict] = None
    pdf_doc: Optional[fitz.Document] = None
    txt_doc: Optional[str] = None


def get_document_handle(path: str) -> DocumentHandle:
    _, extension = os.path.splitext(path)
    handle = get_handle_by_extension(path, extension)
    if handle is not None:
        return handle

    # if no extension, use puremagic to determine the type
    extension = puremagic.from_file(path)
    handle = get_handle_by_extension(path, extension)
    if handle is not None:
        return handle

    raise excs.Error(f'Unrecognized document format: {path}')


def get_handle_by_extension(path: str, extension: str) -> Optional[DocumentHandle]:
    doc_format = ts.DocumentType.DocumentFormat.from_extension(extension)

    try:
        if doc_format == ts.DocumentType.DocumentFormat.HTML:
            return DocumentHandle(doc_format, bs_doc=get_html_handle(path))
        if doc_format == ts.DocumentType.DocumentFormat.MD:
            return DocumentHandle(doc_format, md_ast=get_markdown_handle(path))
        if doc_format == ts.DocumentType.DocumentFormat.PDF:
            return DocumentHandle(doc_format, pdf_doc=get_pdf_handle(path))
        if doc_format == ts.DocumentType.DocumentFormat.XML:
            return DocumentHandle(doc_format, bs_doc=get_xml_handle(path))
        if doc_format == ts.DocumentType.DocumentFormat.TXT:
            return DocumentHandle(doc_format, txt_doc=get_txt(path))
    except Exception as exc:
        raise excs.Error(f'An error occurred processing a {doc_format} document: {path}') from exc

    return None


def get_html_handle(path: str) -> bs4.BeautifulSoup:
    with open(path, 'r', encoding='utf8') as fp:
        doc = bs4.BeautifulSoup(fp, 'lxml')
    if doc.find() is None:
        raise excs.Error(f'Not a valid HTML document: {path}')
    return doc


def get_markdown_handle(path: str) -> dict:
    Env.get().require_package('mistune', [3, 0])
    import mistune

    with open(path, encoding='utf8') as file:
        text = file.read()
    md_ast = mistune.create_markdown(renderer=None)
    return md_ast(text)


def get_pdf_handle(path: str) -> fitz.Document:
    doc = fitz.open(path)
    # check pdf (bc it will work for images)
    if not doc.is_pdf:
        raise excs.Error(f'Not a valid PDF document: {path}')
    # try to read one page
    next(page for page in doc)
    return doc


def get_xml_handle(path: str) -> bs4.BeautifulSoup:
    with open(path, 'r', encoding='utf8') as fp:
        doc = bs4.BeautifulSoup(fp, 'xml')
    if doc.find() is None:
        raise excs.Error(f'Not a valid XML document: {path}')
    return doc


def get_txt(path: str) -> str:
    with open(path, 'r', encoding='utf-8') as fp:
        doc = fp.read()
    return doc



from __future__ import annotations

import abc
import datetime
import enum
import io
import json
import typing
import urllib.parse
import urllib.request
from typing import Any, ClassVar, Iterable, Literal, Mapping, Optional, Sequence, Union

import av
import jsonschema
import jsonschema.protocols
import jsonschema.validators
import numpy as np
import PIL.Image
import pydantic
import sqlalchemy as sql
from typing_extensions import _AnnotatedAlias

import pixeltable.exceptions as excs
from pixeltable.utils import parse_local_file_path

from typing import _GenericAlias  # type: ignore[attr-defined]  # isort: skip


class ColumnType:
    @enum.unique
    class Type(enum.Enum):
        STRING = 0
        INT = 1
        FLOAT = 2
        BOOL = 3
        TIMESTAMP = 4
        JSON = 5
        ARRAY = 6
        IMAGE = 7
        VIDEO = 8
        AUDIO = 9
        DOCUMENT = 10
        DATE = 11

        # exprs that don't evaluate to a computable value in Pixeltable, such as an Image member function
        INVALID = 255

        @classmethod
        def supertype(
            cls,
            type1: Optional['ColumnType.Type'],
            type2: Optional['ColumnType.Type'],
            # we need to pass this in because we can't easily append it as a class member
            common_supertypes: dict[tuple['ColumnType.Type', 'ColumnType.Type'], 'ColumnType.Type'],
        ) -> Optional['ColumnType.Type']:
            if type1 == type2:
                return type1
            t = common_supertypes.get((type1, type2))
            if t is not None:
                return t
            t = common_supertypes.get((type2, type1))
            if t is not None:
                return t
            return None

    @enum.unique
    class DType(enum.Enum):
        """
        Base type used in images and arrays
        """

        BOOL = (0,)
        INT8 = (1,)
        INT16 = (2,)
        INT32 = (3,)
        INT64 = (4,)
        UINT8 = (5,)
        UINT16 = (6,)
        UINT32 = (7,)
        UINT64 = (8,)
        FLOAT16 = (9,)
        FLOAT32 = (10,)
        FLOAT64 = 11

    scalar_types: ClassVar[set[Type]] = {Type.STRING, Type.INT, Type.FLOAT, Type.BOOL, Type.TIMESTAMP, Type.DATE}
    numeric_types: ClassVar[set[Type]] = {Type.INT, Type.FLOAT}
    common_supertypes: ClassVar[dict[tuple[Type, Type], Type]] = {
        (Type.BOOL, Type.INT): Type.INT,
        (Type.BOOL, Type.FLOAT): Type.FLOAT,
        (Type.INT, Type.FLOAT): Type.FLOAT,
    }

    def __init__(self, t: Type, nullable: bool = False):
        self._type = t
        self._nullable = nullable

    def has_supertype(self) -> bool:
        return True

    @property
    def nullable(self) -> bool:
        return self._nullable

    @property
    def type_enum(self) -> Type:
        return self._type

    def serialize(self) -> str:
        return json.dumps(self.as_dict())

    def copy(self, nullable: bool) -> ColumnType:
        # Default implementation calls unary initializer
        if nullable == self.nullable:
            return self
        else:
            return self.__class__(nullable=nullable)  # type: ignore[call-arg]

    @classmethod
    def serialize_list(cls, type_list: list[ColumnType]) -> str:
        return json.dumps([t.as_dict() for t in type_list])

    def as_dict(self) -> dict:
        return {'_classname': self.__class__.__name__, **self._as_dict()}

    def _as_dict(self) -> dict:
        return {'nullable': self.nullable}

    @classmethod
    def deserialize(cls, type_str: str) -> ColumnType:
        type_dict = json.loads(type_str)
        return cls.from_dict(type_dict)

    @classmethod
    def deserialize_list(cls, type_list_str: str) -> list[ColumnType]:
        type_dict_list = json.loads(type_list_str)
        return [cls.from_dict(type_dict) for type_dict in type_dict_list]

    @classmethod
    def from_dict(cls, type_dict: dict) -> ColumnType:
        assert '_classname' in type_dict
        type_class = globals()[type_dict['_classname']]
        return type_class._from_dict(type_dict)

    @classmethod
    def _from_dict(cls, d: dict) -> ColumnType:
        """
        Default implementation: simply invoke c'tor
        """
        assert 'nullable' in d
        return cls(nullable=d['nullable'])  # type: ignore[call-arg]

    @classmethod
    def make_type(cls, t: Type) -> ColumnType:
        assert t != cls.Type.INVALID
        if t == cls.Type.STRING:
            return StringType()
        if t == cls.Type.INT:
            return IntType()
        if t == cls.Type.FLOAT:
            return FloatType()
        if t == cls.Type.BOOL:
            return BoolType()
        if t == cls.Type.TIMESTAMP:
            return TimestampType()
        if t == cls.Type.ARRAY:
            return ArrayType()
        if t == cls.Type.JSON:
            return JsonType()
        if t == cls.Type.IMAGE:
            return ImageType()
        if t == cls.Type.VIDEO:
            return VideoType()
        if t == cls.Type.AUDIO:
            return AudioType()
        if t == cls.Type.DOCUMENT:
            return DocumentType()
        if t == cls.Type.DATE:
            return DateType()

    def __repr__(self) -> str:
        return self._to_str(as_schema=False)

    def _to_str(self, as_schema: bool) -> str:
        base_str = self._to_base_str()
        if as_schema:
            return base_str if self.nullable else f'Required[{base_str}]'
        else:
            return f'Optional[{base_str}]' if self.nullable else base_str

    def _to_base_str(self) -> str:
        """
        String representation of this type, disregarding nullability. Default implementation is to camel-case
        the type name; subclasses may override.
        """
        return self._type.name[0] + self._type.name[1:].lower()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ColumnType) and self.matches(other) and self.nullable == other.nullable

    def __hash__(self) -> int:
        return hash((self._type, self.nullable))

    def is_supertype_of(self, other: ColumnType, ignore_nullable: bool = False) -> bool:
        if ignore_nullable:
            supertype = self.supertype(other)
            if supertype is None:
                return False
            return supertype.matches(self)
        else:
            return self.supertype(other) == self

    def matches(self, other: ColumnType) -> bool:
        """Two types match if they're equal, aside from nullability"""
        # Default: just compare base types (this works for all types whose only parameter is nullable)
        return self._type == other._type

    def supertype(self, other: ColumnType) -> Optional[ColumnType]:
        if self == other:
            return self
        if self.matches(other):
            return self.copy(nullable=(self.nullable or other.nullable))

        if self.is_invalid_type():
            return other.copy(nullable=(self.nullable or other.nullable))
        if other.is_invalid_type():
            return self.copy(nullable=(self.nullable or other.nullable))

        if self.is_scalar_type() and other.is_scalar_type():
            t = self.Type.supertype(self._type, other._type, self.common_supertypes)
            if t is not None:
                return self.make_type(t).copy(nullable=(self.nullable or other.nullable))
            return None

        return None

    @classmethod
    def infer_literal_type(cls, val: Any, nullable: bool = False) -> Optional[ColumnType]:
        if val is None:
            return InvalidType(nullable=True)
        if isinstance(val, str):
            return StringType(nullable=nullable)
        if isinstance(val, bool):
            # We have to check bool before int, because isinstance(b, int) is True if b is a Python bool
            return BoolType(nullable=nullable)
        if isinstance(val, int):
            return IntType(nullable=nullable)
        if isinstance(val, float):
            return FloatType(nullable=nullable)
        # When checking types of dates / timestamps, be aware that a datetime is also a date,
        # but a date is not a datetime. So check for datetime first.
        if isinstance(val, datetime.datetime):
            return TimestampType(nullable=nullable)
        if isinstance(val, datetime.date):
            return DateType(nullable=nullable)
        if isinstance(val, PIL.Image.Image):
            return ImageType(width=val.width, height=val.height, mode=val.mode, nullable=nullable)
        if isinstance(val, np.ndarray):
            col_type = ArrayType.from_literal(val, nullable=nullable)
            if col_type is not None:
                return col_type
        # this could still be json-serializable
        if isinstance(val, (list, tuple, dict, np.ndarray, pydantic.BaseModel)):
            try:
                JsonType().validate_literal(val)
                return JsonType(nullable=nullable)
            except TypeError:
                return None
        return None

    @classmethod
    def infer_common_literal_type(cls, vals: Iterable[Any]) -> Optional[ColumnType]:
        """
        Returns the most specific type that is a supertype of all literals in `vals`. If no such type
        exists, returns None.

        Args:
            vals: A collection of literals.
        """
        inferred_type: Optional[ColumnType] = None
        for val in vals:
            val_type = cls.infer_literal_type(val)
            if inferred_type is None:
                inferred_type = val_type
            else:
                inferred_type = inferred_type.supertype(val_type)
            if inferred_type is None:
                return None
            if not inferred_type.has_supertype():
                return inferred_type
        return inferred_type

    @classmethod
    def from_python_type(
        cls, t: Union[type, _GenericAlias], nullable_default: bool = False, allow_builtin_types: bool = True
    ) -> Optional[ColumnType]:
        """
        Convert a Python type into a Pixeltable `ColumnType` instance.

        Args:
            t: The Python type.
            nullable_default: If True, then the returned `ColumnType` will be nullable unless it is marked as
                `Required`.
            allow_builtin_types: If True, then built-in types such as `str`, `int`, `float`, etc., will be
                allowed (as in UDF definitions). If False, then only Pixeltable types such as `pxt.String`,
                `pxt.Int`, etc., will be allowed (as in schema definitions). `Optional` and `Required`
                designations will be allowed regardless.
        """
        origin = typing.get_origin(t)
        type_args = typing.get_args(t)
        if origin is typing.Union:
            # Check if `t` has the form Optional[T].
            if len(type_args) == 2 and type(None) in type_args:
                # `t` is a type of the form Optional[T] (equivalently, Union[T, None] or Union[None, T]).
                # We treat it as the underlying type but with nullable=True.
                underlying_py_type = type_args[0] if type_args[1] is type(None) else type_args[1]
                underlying = cls.from_python_type(underlying_py_type, allow_builtin_types=allow_builtin_types)
                if underlying is not None:
                    return underlying.copy(nullable=True)
        elif origin is Required:
            assert len(type_args) == 1
            return cls.from_python_type(
                type_args[0], nullable_default=False, allow_builtin_types=allow_builtin_types
            ).copy(nullable=False)
        elif origin is typing.Annotated:
            origin = type_args[0]
            parameters = type_args[1]
            if isinstance(parameters, ColumnType):
                return parameters.copy(nullable=nullable_default)
        else:
            # It's something other than Optional[T], Required[T], or an explicitly annotated type.
            if origin is not None:
                # Discard type parameters to ensure that parameterized types such as `list[T]`
                # are correctly mapped to Pixeltable types.
                t = origin
            if isinstance(t, type) and issubclass(t, _PxtType):
                return t.as_col_type(nullable=nullable_default)
            elif allow_builtin_types:
                if t is Literal and len(type_args) > 0:
                    literal_type = cls.infer_common_literal_type(type_args)
                    if literal_type is None:
                        return None
                    return literal_type.copy(nullable=(literal_type.nullable or nullable_default))
                if t is str:
                    return StringType(nullable=nullable_default)
                if t is int:
                    return IntType(nullable=nullable_default)
                if t is float:
                    return FloatType(nullable=nullable_default)
                if t is bool:
                    return BoolType(nullable=nullable_default)
                if t is datetime.datetime:
                    return TimestampType(nullable=nullable_default)
                if t is datetime.date:
                    return DateType(nullable=nullable_default)
                if t is PIL.Image.Image:
                    return ImageType(nullable=nullable_default)
                if isinstance(t, type) and issubclass(t, (Sequence, Mapping, pydantic.BaseModel)):
                    return JsonType(nullable=nullable_default)
        return None

    @classmethod
    def normalize_type(
        cls,
        t: Union[ColumnType, type, _AnnotatedAlias],
        nullable_default: bool = False,
        allow_builtin_types: bool = True,
    ) -> ColumnType:
        """
        Convert any type recognizable by Pixeltable to its corresponding ColumnType.
        """
        if isinstance(t, ColumnType):
            return t
        col_type = cls.from_python_type(t, nullable_default, allow_builtin_types)
        if col_type is None:
            cls.__raise_exc_for_invalid_type(t)
        return col_type

    __TYPE_SUGGESTIONS: ClassVar[list[tuple[type, str]]] = [
        (str, 'pxt.String'),
        (bool, 'pxt.Bool'),
        (int, 'pxt.Int'),
        (float, 'pxt.Float'),
        (datetime.datetime, 'pxt.Timestamp'),
        (datetime.date, 'pxt.Date'),
        (PIL.Image.Image, 'pxt.Image'),
        (Sequence, 'pxt.Json'),
        (Mapping, 'pxt.Json'),
    ]

    @classmethod
    def __raise_exc_for_invalid_type(cls, t: Union[type, _AnnotatedAlias]) -> None:
        for builtin_type, suggestion in cls.__TYPE_SUGGESTIONS:
            if t is builtin_type or (isinstance(t, type) and issubclass(t, builtin_type)):
                name = t.__name__ if t.__module__ == 'builtins' else f'{t.__module__}.{t.__name__}'
                raise excs.Error(f'Standard Python type `{name}` cannot be used here; use `{suggestion}` instead')
        raise excs.Error(f'Unknown type: {t}')

    @classmethod
    def from_json_schema(cls, schema: dict[str, Any]) -> Optional[ColumnType]:
        # We first express the JSON schema as a Python type, and then convert it to a Pixeltable type.
        # TODO: Is there a meaningful fallback if one of these operations fails? (Maybe another use case for a pxt Any
        #     type?)
        py_type = cls.__json_schema_to_py_type(schema)
        return cls.from_python_type(py_type) if py_type is not None else None

    @classmethod
    def __json_schema_to_py_type(cls, schema: dict[str, Any]) -> Union[type, _GenericAlias, None]:
        if 'type' in schema:
            if schema['type'] == 'null':
                return type(None)
            if schema['type'] == 'string':
                return str
            if schema['type'] == 'integer':
                return int
            if schema['type'] == 'number':
                return float
            if schema['type'] == 'boolean':
                return bool
            if schema['type'] in ('array', 'object'):
                return list
        elif 'anyOf' in schema:
            subscripts = tuple(cls.__json_schema_to_py_type(subschema) for subschema in schema['anyOf'])
            if all(subscript is not None for subscript in subscripts):
                return Union[subscripts]

        return None

    def validate_literal(self, val: Any) -> None:
        """Raise TypeError if val is not a valid literal for this type"""
        if val is None:
            if not self.nullable:
                raise TypeError('Expected non-None value')
            else:
                return
        self._validate_literal(val)

    def validate_media(self, val: Any) -> None:
        """
        Raise TypeError if val is not a path to a valid media file (or a valid in-memory byte sequence) for this type
        """
        if self.is_media_type():
            raise NotImplementedError(f'validate_media() not implemented for {self.__class__.__name__}')

    def _validate_file_path(self, val: Any) -> None:
        """Raises TypeError if not a valid local file path or not a path/byte sequence"""
        if isinstance(val, str):
            path = parse_local_file_path(val)
            if path is not None and not path.is_file():
                raise TypeError(f'File not found: {path}')
        elif not isinstance(val, bytes):
            raise TypeError(f'expected file path or bytes, got {type(val)}')

    @abc.abstractmethod
    def _validate_literal(self, val: Any) -> None:
        """Raise TypeError if val is not a valid literal for this type"""

    def _create_literal(self, val: Any) -> Any:
        """Create a literal of this type from val, including any needed conversions.
        val is guaranteed to be non-None"""
        return val

    def create_literal(self, val: Any) -> Any:
        """Create a literal of this type from val or raise TypeError if not possible"""
        if val is not None:
            val = self._create_literal(val)

        self.validate_literal(val)
        return val

    def print_value(self, val: Any) -> str:
        return str(val)

    def is_scalar_type(self) -> bool:
        return self._type in self.scalar_types

    def is_numeric_type(self) -> bool:
        return self._type in self.numeric_types

    def is_invalid_type(self) -> bool:
        return self._type == self.Type.INVALID

    def is_string_type(self) -> bool:
        return self._type == self.Type.STRING

    def is_int_type(self) -> bool:
        return self._type == self.Type.INT

    def is_float_type(self) -> bool:
        return self._type == self.Type.FLOAT

    def is_bool_type(self) -> bool:
        return self._type == self.Type.BOOL

    def is_timestamp_type(self) -> bool:
        return self._type == self.Type.TIMESTAMP

    def is_date_type(self) -> bool:
        return self._type == self.Type.DATE

    def is_json_type(self) -> bool:
        return self._type == self.Type.JSON

    def is_array_type(self) -> bool:
        return self._type == self.Type.ARRAY

    def is_image_type(self) -> bool:
        return self._type == self.Type.IMAGE

    def is_video_type(self) -> bool:
        return self._type == self.Type.VIDEO

    def is_audio_type(self) -> bool:
        return self._type == self.Type.AUDIO

    def is_document_type(self) -> bool:
        return self._type == self.Type.DOCUMENT

    def is_media_type(self) -> bool:
        # types that refer to external media files
        return self.is_image_type() or self.is_video_type() or self.is_audio_type() or self.is_document_type()

    @classmethod
    @abc.abstractmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        """
        Return corresponding SQLAlchemy type.
        """

    def to_json_schema(self) -> dict[str, Any]:
        if self.nullable:
            return {'anyOf': [self._to_json_schema(), {'type': 'null'}]}
        else:
            return self._to_json_schema()

    def _to_json_schema(self) -> dict[str, Any]:
        raise excs.Error(f'Pixeltable type {self} is not a valid JSON type')


class InvalidType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.INVALID, nullable=nullable)

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.types.NullType()

    def print_value(self, val: Any) -> str:
        return str(val)

    def _validate_literal(self, val: Any) -> None:
        raise AssertionError()


class StringType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.STRING, nullable=nullable)

    def has_supertype(self) -> bool:
        return not self.nullable

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.String()

    def _to_json_schema(self) -> dict[str, Any]:
        return {'type': 'string'}

    def print_value(self, val: Any) -> str:
        return f"'{val}'"

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, str):
            raise TypeError(f'Expected string, got {val.__class__.__name__}')

    def _create_literal(self, val: Any) -> Any:
        # Replace null byte within python string with space to avoid issues with Postgres.
        # Use a space to avoid merging words.
        # TODO(orm): this will also be an issue with JSON inputs, would space still be a good replacement?
        if isinstance(val, str) and '\x00' in val:
            return val.replace('\x00', ' ')
        return val


class IntType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.INT, nullable=nullable)

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.BigInteger()

    def _to_json_schema(self) -> dict[str, Any]:
        return {'type': 'integer'}

    def _validate_literal(self, val: Any) -> None:
        # bool is a subclass of int, so we need to check for it
        # explicitly first
        if isinstance(val, bool) or not isinstance(val, int):
            raise TypeError(f'Expected int, got {val.__class__.__name__}')


class FloatType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.FLOAT, nullable=nullable)

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.Float()

    def _to_json_schema(self) -> dict[str, Any]:
        return {'type': 'number'}

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, float):
            raise TypeError(f'Expected float, got {val.__class__.__name__}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, int):
            return float(val)
        return val


class BoolType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.BOOL, nullable=nullable)

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.Boolean()

    def _to_json_schema(self) -> dict[str, Any]:
        return {'type': 'boolean'}

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, bool):
            raise TypeError(f'Expected bool, got {val.__class__.__name__}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, int):
            return bool(val)
        return val


class TimestampType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.TIMESTAMP, nullable=nullable)

    def has_supertype(self) -> bool:
        return not self.nullable

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.TIMESTAMP(timezone=True)

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, datetime.datetime):
            raise TypeError(f'Expected datetime.datetime, got {val.__class__.__name__}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, str):
            return datetime.datetime.fromisoformat(val)
        if isinstance(val, datetime.datetime):
            return val
        return val


class DateType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.DATE, nullable=nullable)

    def has_supertype(self) -> bool:
        return not self.nullable

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.Date()

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, datetime.date):
            raise TypeError(f'Expected datetime.date, got {val.__class__.__name__}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, str):
            return datetime.datetime.fromisoformat(val).date()
        if isinstance(val, datetime.date):
            return val
        return val


class JsonType(ColumnType):
    json_schema: Optional[dict[str, Any]]
    __validator: Optional[jsonschema.protocols.Validator]

    def __init__(self, json_schema: Optional[dict[str, Any]] = None, nullable: bool = False):
        super().__init__(self.Type.JSON, nullable=nullable)
        self.json_schema = json_schema
        if json_schema is None:
            self.__validator = None
        else:
            validator_cls = jsonschema.validators.validator_for(json_schema)
            validator_cls.check_schema(json_schema)
            self.__validator = validator_cls(json_schema)

    def copy(self, nullable: bool) -> ColumnType:
        return JsonType(json_schema=self.json_schema, nullable=nullable)

    def matches(self, other: ColumnType) -> bool:
        return isinstance(other, JsonType) and self.json_schema == other.json_schema

    def _as_dict(self) -> dict:
        result = super()._as_dict()
        if self.json_schema is not None:
            result.update({'json_schema': self.json_schema})
        return result

    @classmethod
    def _from_dict(cls, d: dict) -> ColumnType:
        return cls(json_schema=d.get('json_schema'), nullable=d['nullable'])

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.dialects.postgresql.JSONB()

    def _to_json_schema(self) -> dict[str, Any]:
        if self.json_schema is None:
            return {}
        return self.json_schema

    def print_value(self, val: Any) -> str:
        val_type = self.infer_literal_type(val)
        if val_type is None:
            return super().print_value(val)
        if val_type == self:
            return str(val)
        return val_type.print_value(val)

    def _validate_literal(self, val: Any) -> None:
        if isinstance(val, tuple):
            val = list(val)
        if isinstance(val, pydantic.BaseModel):
            val = val.model_dump()
        if not self.__is_valid_json(val):
            raise TypeError(f'That literal is not a valid Pixeltable JSON object: {val}')
        if self.__validator is not None:
            self.__validator.validate(val)

    @classmethod
    def __is_valid_json(cls, val: Any) -> bool:
        if val is None or isinstance(val, (str, int, float, bool)):
            return True
        if isinstance(val, (list, tuple)):
            return all(cls.__is_valid_json(v) for v in val)
        if isinstance(val, dict):
            return all(isinstance(k, str) and cls.__is_valid_json(v) for k, v in val.items())
        return False

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, tuple):
            val = list(val)
        if isinstance(val, pydantic.BaseModel):
            return val.model_dump()
        return val

    def supertype(self, other: ColumnType) -> Optional[JsonType]:
        # Try using the (much faster) supertype logic in ColumnType first. That will work if, for example, the types
        # are identical except for nullability. If that doesn't work and both types are JsonType, then we will need to
        # merge their schemas.
        basic_supertype = super().supertype(other)
        if basic_supertype is not None:
            assert isinstance(basic_supertype, JsonType)
            return basic_supertype

        if not isinstance(other, JsonType):
            return None

        if self.json_schema is None or other.json_schema is None:
            return JsonType(nullable=(self.nullable or other.nullable))

        superschema = self.__superschema(self.json_schema, other.json_schema)

        return JsonType(
            json_schema=(None if len(superschema) == 0 else superschema), nullable=(self.nullable or other.nullable)
        )

    @classmethod
    def __superschema(cls, a: dict[str, Any], b: dict[str, Any]) -> Optional[dict[str, Any]]:
        # Defining a general type hierarchy over all JSON schemas would be a challenging problem. In order to keep
        # things manageable, we only define a hierarchy among "conforming" schemas, which provides enough generality
        # for the most important use cases (unions for type inference, validation of inline exprs). A schema is
        # considered to be conforming if either:
        # (i) it is a scalar (string, integer, number, boolean) or dictionary (object) type; or
        # (ii) it is an "anyOf" schema of one of the above types and the exact schema {'type': 'null'}.
        # Conforming schemas are organized into a type hierarchy in an internally consistent way. Nonconforming
        # schemas are allowed, but they are isolates in the type hierarchy: a nonconforming schema has no proper
        # subtypes, and its only proper supertype is an unconstrained JsonType().
        #
        # There is some subtlety in the handling of nullable fields. Nullable fields are represented in JSON
        # schemas as (for example) {'anyOf': [{'type': 'string'}, {'type': 'null'}]}. When finding the supertype
        # of schemas that might be nullable, we first unpack the 'anyOf's, find the supertype of the underlyings,
        # then reapply the 'anyOf' if appropriate. The top-level schema (i.e., JsonType.json_schema) is presumed
        # to NOT be in this form (since nullability is indicated by the `nullable` field of the JsonType object),
        # so this subtlety is applicable only to types that occur in subfields.
        #
        # There is currently no special handling of lists; distinct schemas with type 'array' will union to the
        # generic {'type': 'array'} schema. This could be a TODO item if there is a need for it in the future.

        if a == b:
            return a

        if 'properties' in a and 'properties' in b:
            a_props = a['properties']
            b_props = b['properties']
            a_req = a.get('required', [])
            b_req = b.get('required', [])
            super_props = {}
            super_req = []
            for key, a_prop_schema in a_props.items():
                if key in b_props:  # in both a and b
                    prop_schema = cls.__superschema_with_nulls(a_prop_schema, b_props[key])
                    super_props[key] = prop_schema
                    if key in a_req and key in b_req:
                        super_req.append(key)
                else:  # in a but not b
                    # Add it to the supertype schema as optional (regardless of its status in a)
                    super_props[key] = a_prop_schema
            for key, b_prop_schema in b_props.items():
                if key not in a_props:  # in b but not a
                    super_props[key] = b_prop_schema
            schema = {'type': 'object', 'properties': super_props}
            if len(super_req) > 0:
                schema['required'] = super_req
            return schema

        a_type = a.get('type')
        b_type = b.get('type')

        if a_type in ('string', 'integer', 'number', 'boolean', 'object', 'array') and a_type == b_type:
            # a and b both have the same type designation, but are not identical. This can happen if
            # (for example) they have validators or other attributes that differ. In this case, we
            # generalize to {'type': t}, where t is their shared type, with no other qualifications.
            return {'type': a_type}

        return {}  # Unresolvable type conflict; the supertype is an unrestricted JsonType.

    @classmethod
    def __superschema_with_nulls(cls, a: dict[str, Any], b: dict[str, Any]) -> Optional[dict[str, Any]]:
        a, a_nullable = cls.__unpack_null_from_schema(a)
        b, b_nullable = cls.__unpack_null_from_schema(b)

        result = cls.__superschema(a, b)
        if len(result) > 0 and (a_nullable or b_nullable):
            # if len(result) == 0, then null is implicitly accepted; otherwise, we need to explicitly allow it
            return {'anyOf': [result, {'type': 'null'}]}
        return result

    @classmethod
    def __unpack_null_from_schema(cls, s: dict[str, Any]) -> tuple[dict[str, Any], bool]:
        if 'anyOf' in s and len(s['anyOf']) == 2 and {'type': 'null'} in s['anyOf']:
            try:
                return next(s for s in s['anyOf'] if s != {'type': 'null'}), True
            except StopIteration:
                pass
        return s, False

    def _to_base_str(self) -> str:
        if self.json_schema is None:
            return 'Json'
        elif 'title' in self.json_schema:
            return f'Json[{self.json_schema["title"]}]'
        else:
            return f'Json[{self.json_schema}]'


class ArrayType(ColumnType):
    shape: Optional[tuple[Optional[int], ...]]
    pxt_dtype: Optional[ColumnType]
    dtype: Optional[ColumnType.Type]

    def __init__(
        self,
        shape: Optional[tuple[Optional[int], ...]] = None,
        dtype: Optional[ColumnType] = None,
        nullable: bool = False,
    ):
        super().__init__(self.Type.ARRAY, nullable=nullable)
        assert shape is None or dtype is not None, (shape, dtype)  # cannot specify a shape without a dtype
        assert (
            dtype is None
            or dtype.is_int_type()
            or dtype.is_float_type()
            or dtype.is_bool_type()
            or dtype.is_string_type()
        )

        self.shape = shape
        self.pxt_dtype = dtype  # we need this for copy() and __str__()
        self.dtype = None if dtype is None else dtype._type

    def copy(self, nullable: bool) -> ColumnType:
        return ArrayType(self.shape, self.pxt_dtype, nullable=nullable)

    def matches(self, other: ColumnType) -> bool:
        return isinstance(other, ArrayType) and self.shape == other.shape and self.dtype == other.dtype

    def __hash__(self) -> int:
        return hash((self._type, self.nullable, self.shape, self.dtype))

    def supertype(self, other: ColumnType) -> Optional[ArrayType]:
        basic_supertype = super().supertype(other)
        if basic_supertype is not None:
            assert isinstance(basic_supertype, ArrayType)
            return basic_supertype

        if not isinstance(other, ArrayType):
            return None

        super_dtype = self.Type.supertype(self.dtype, other.dtype, self.common_supertypes)
        if super_dtype is None:
            # if the dtypes are incompatible, then the supertype is a fully general array
            return ArrayType(nullable=(self.nullable or other.nullable))
        super_shape: Optional[tuple[Optional[int], ...]]
        if self.shape is None or other.shape is None or len(self.shape) != len(other.shape):
            super_shape = None
        else:
            super_shape = tuple(n1 if n1 == n2 else None for n1, n2 in zip(self.shape, other.shape))
        return ArrayType(super_shape, self.make_type(super_dtype), nullable=(self.nullable or other.nullable))

    def _as_dict(self) -> dict:
        result = super()._as_dict()
        shape_as_list = None if self.shape is None else list(self.shape)
        dtype_value = None if self.dtype is None else self.dtype.value
        result.update(shape=shape_as_list, dtype=dtype_value)
        return result

    def _to_base_str(self) -> str:
        if self.shape is None and self.dtype is None:
            return 'Array'
        if self.shape is None:
            return f'Array[{self.pxt_dtype}]'
        assert self.dtype is not None
        return f'Array[{self.shape}, {self.pxt_dtype}]'

    @classmethod
    def _from_dict(cls, d: dict) -> ColumnType:
        assert 'shape' in d
        assert 'dtype' in d
        shape = None if d['shape'] is None else tuple(d['shape'])
        dtype = None if d['dtype'] is None else cls.make_type(cls.Type(d['dtype']))
        return cls(shape, dtype, nullable=d['nullable'])

    @classmethod
    def from_np_dtype(cls, dtype: np.dtype, nullable: bool) -> Optional[ColumnType]:
        """
        Return pixeltable type corresponding to a given simple numpy dtype
        """
        if np.issubdtype(dtype, np.integer):
            return IntType(nullable=nullable)

        if np.issubdtype(dtype, np.floating):
            return FloatType(nullable=nullable)

        if dtype == np.bool_:
            return BoolType(nullable=nullable)

        if np.issubdtype(dtype, np.str_):
            return StringType(nullable=nullable)

        if np.issubdtype(dtype, np.character):
            return StringType(nullable=nullable)

        if np.issubdtype(dtype, np.datetime64):
            unit, _ = np.datetime_data(dtype)
            if unit in ['D', 'M', 'Y']:
                return DateType(nullable=nullable)
            else:
                return TimestampType(nullable=nullable)

        return None

    @classmethod
    def from_literal(cls, val: np.ndarray, nullable: bool = False) -> Optional[ArrayType]:
        # determine our dtype
        assert isinstance(val, np.ndarray)
        pxttype: Optional[ColumnType] = cls.from_np_dtype(val.dtype, nullable)
        if pxttype is None:
            return None
        return cls(val.shape, dtype=pxttype, nullable=nullable)

    def is_valid_literal(self, val: np.ndarray) -> bool:
        if not isinstance(val, np.ndarray):
            return False

        # If a dtype is specified, check that there's a match
        if self.dtype is not None and not np.issubdtype(val.dtype, self.numpy_dtype()):
            return False

        # If no dtype is specified, we still need to check that the dtype is one of the supported types
        if self.dtype is None and not any(
            np.issubdtype(val.dtype, ndtype) for ndtype in [np.int64, np.float32, np.bool_, np.str_]
        ):
            return False

        # If a shape is specified, check that there's a match
        if self.shape is not None:
            if len(val.shape) != len(self.shape):
                return False
            # check that the shapes are compatible
            for n1, n2 in zip(val.shape, self.shape):
                assert n1 is not None  # `val` must have a concrete shape
                if n2 is None:
                    continue  # wildcard
                if n1 != n2:
                    return False

        return True

    def _to_json_schema(self) -> dict[str, Any]:
        return {'type': 'array', 'items': self.pxt_dtype._to_json_schema()}

    def _validate_literal(self, val: Any) -> None:
        if not isinstance(val, np.ndarray):
            raise TypeError(f'Expected numpy.ndarray, got {val.__class__.__name__}')
        if not self.is_valid_literal(val):
            if self.shape is not None:
                raise TypeError(
                    f'Expected numpy.ndarray({self.shape}, dtype={self.numpy_dtype()}), '
                    f'got numpy.ndarray({val.shape}, dtype={val.dtype})'
                )
            elif self.dtype is not None:
                raise TypeError(
                    f'Expected numpy.ndarray of dtype {self.numpy_dtype()}, got numpy.ndarray of dtype {val.dtype}'
                )
            else:
                raise TypeError(f'Unsupported dtype for numpy.ndarray: {val.dtype}')

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, (list, tuple)):
            # map python float to whichever numpy float is
            # declared for this type, rather than assume float64
            return np.array(val, dtype=self.numpy_dtype())
        return val

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.LargeBinary()

    def numpy_dtype(self) -> Optional[np.dtype]:
        if self.dtype is None:
            return None
        if self.dtype == self.Type.INT:
            return np.dtype(np.int64)
        if self.dtype == self.Type.FLOAT:
            return np.dtype(np.float32)
        if self.dtype == self.Type.BOOL:
            return np.dtype(np.bool_)
        if self.dtype == self.Type.STRING:
            return np.dtype(np.str_)
        raise AssertionError(self.dtype)


class ImageType(ColumnType):
    def __init__(
        self,
        width: Optional[int] = None,
        height: Optional[int] = None,
        size: Optional[tuple[int, int]] = None,
        mode: Optional[str] = None,
        nullable: bool = False,
    ):
        """
        TODO: does it make sense to specify only width or height?
        """
        super().__init__(self.Type.IMAGE, nullable=nullable)
        assert not (width is not None and size is not None)
        assert not (height is not None and size is not None)
        if size is not None:
            self.width = size[0]
            self.height = size[1]
        else:
            self.width = width
            self.height = height
        self.mode = mode

    def copy(self, nullable: bool) -> ColumnType:
        return ImageType(self.width, self.height, mode=self.mode, nullable=nullable)

    def _to_base_str(self) -> str:
        params = []
        if self.width is not None or self.height is not None:
            params.append(f'({self.width}, {self.height})')
        if self.mode is not None:
            params.append(repr(self.mode))
        if len(params) == 0:
            params_str = ''
        else:
            params_str = f'[{", ".join(params)}]'
        return f'Image{params_str}'

    def matches(self, other: ColumnType) -> bool:
        return (
            isinstance(other, ImageType)
            and self.width == other.width
            and self.height == other.height
            and self.mode == other.mode
        )

    def __hash__(self) -> int:
        return hash((self._type, self.nullable, self.size, self.mode))

    def supertype(self, other: ColumnType) -> Optional[ImageType]:
        basic_supertype = super().supertype(other)
        if basic_supertype is not None:
            assert isinstance(basic_supertype, ImageType)
            return basic_supertype

        if not isinstance(other, ImageType):
            return None

        width = self.width if self.width == other.width else None
        height = self.height if self.height == other.height else None
        mode = self.mode if self.mode == other.mode else None
        return ImageType(width=width, height=height, mode=mode, nullable=(self.nullable or other.nullable))

    @property
    def size(self) -> Optional[tuple[int, int]]:
        if self.width is None or self.height is None:
            return None
        return (self.width, self.height)

    def _as_dict(self) -> dict:
        result = super()._as_dict()
        result.update(width=self.width, height=self.height, mode=self.mode)
        return result

    @classmethod
    def _from_dict(cls, d: dict) -> ColumnType:
        assert 'width' in d
        assert 'height' in d
        assert 'mode' in d
        return cls(width=d['width'], height=d['height'], mode=d['mode'], nullable=d['nullable'])

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        return sql.String()

    def _create_literal(self, val: Any) -> Any:
        if isinstance(val, str) and val.startswith('data:'):
            # try parsing this as a `data:` URL, and if successful, decode the image immediately
            try:
                with urllib.request.urlopen(val) as response:
                    b = response.read()
                img = PIL.Image.open(io.BytesIO(b))
                img.load()
                return img
            except Exception as exc:
                error_msg_val = val if len(val) < 50 else val[:50] + '...'
                raise excs.Error(f'data URL could not be decoded into a valid image: {error_msg_val}') from exc
        return val

    def _validate_literal(self, val: Any) -> None:
        if isinstance(val, PIL.Image.Image):
            return
        self._validate_file_path(val)

    def validate_media(self, val: Any) -> None:
        assert isinstance(val, str)
        try:
            _ = PIL.Image.open(val)
        except PIL.UnidentifiedImageError:
            raise excs.Error(f'Not a valid image: {val}') from None


class VideoType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.VIDEO, nullable=nullable)

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        # stored as a file path
        return sql.String()

    def _validate_literal(self, val: Any) -> None:
        self._validate_file_path(val)

    def validate_media(self, val: Any) -> None:
        assert isinstance(val, str)
        try:
            with av.open(val, 'r') as fh:
                if len(fh.streams.video) == 0:
                    raise excs.Error(f'Not a valid video: {val}')
                # decode a few frames to make sure it's playable
                # TODO: decode all frames? but that's very slow
                num_decoded = 0
                for frame in fh.decode(video=0):
                    _ = frame.to_image()
                    num_decoded += 1
                    if num_decoded == 10:
                        break
                if num_decoded < 2:
                    # this is most likely an image file
                    raise excs.Error(f'Not a valid video: {val}')
        except av.FFmpegError:
            raise excs.Error(f'Not a valid video: {val}') from None


class AudioType(ColumnType):
    def __init__(self, nullable: bool = False):
        super().__init__(self.Type.AUDIO, nullable=nullable)

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        # stored as a file path
        return sql.String()

    def _validate_literal(self, val: Any) -> None:
        self._validate_file_path(val)

    def validate_media(self, val: Any) -> None:
        try:
            with av.open(val) as container:
                if len(container.streams.audio) == 0:
                    raise excs.Error(f'No audio stream in file: {val}')
                audio_stream = container.streams.audio[0]

                # decode everything to make sure it's playable
                # TODO: is there some way to verify it's a playable audio file other than decoding all of it?
                for packet in container.demux(audio_stream):
                    for _ in packet.decode():
                        pass
        except av.FFmpegError as e:
            raise excs.Error(f'Not a valid audio file: {val}\n{e}') from None


class DocumentType(ColumnType):
    @enum.unique
    class DocumentFormat(enum.Enum):
        HTML = 0
        MD = 1
        PDF = 2
        XML = 3
        TXT = 4

        @classmethod
        def from_extension(cls, ext: str) -> Optional['DocumentType.DocumentFormat']:
            if ext in ('.htm', '.html'):
                return cls.HTML
            if ext == '.md':
                return cls.MD
            if ext == '.pdf':
                return cls.PDF
            if ext == '.xml':
                return cls.XML
            if ext == '.txt':
                return cls.TXT
            return None

    def __init__(self, nullable: bool = False, doc_formats: Optional[str] = None):
        super().__init__(self.Type.DOCUMENT, nullable=nullable)
        self.doc_formats = doc_formats
        if doc_formats is not None:
            type_strs = doc_formats.split(',')
            for type_str in type_strs:
                if not hasattr(self.DocumentFormat, type_str):
                    raise ValueError(f'Invalid document type: {type_str}')
            self._doc_formats = [self.DocumentFormat[type_str.upper()] for type_str in type_strs]
        else:
            self._doc_formats = list(self.DocumentFormat)

    def copy(self, nullable: bool) -> ColumnType:
        return DocumentType(doc_formats=self.doc_formats, nullable=nullable)

    def matches(self, other: ColumnType) -> bool:
        return isinstance(other, DocumentType) and self._doc_formats == other._doc_formats

    def __hash__(self) -> int:
        return hash((self._type, self.nullable, self._doc_formats))

    @classmethod
    def to_sa_type(cls) -> sql.types.TypeEngine:
        # stored as a file path
        return sql.String()

    def _validate_literal(self, val: Any) -> None:
        self._validate_file_path(val)

    def validate_media(self, val: Any) -> None:
        assert isinstance(val, str)
        from pixeltable.utils.documents import get_document_handle

        _ = get_document_handle(val)


T = typing.TypeVar('T')


class Required(typing.Generic[T]):
    """
    Marker class to indicate that a column is non-nullable in a schema definition. This has no meaning as a type hint,
    and is intended only for schema declarations.
    """

    pass


String = typing.Annotated[str, StringType(nullable=False)]
Int = typing.Annotated[int, IntType(nullable=False)]
Float = typing.Annotated[float, FloatType(nullable=False)]
Bool = typing.Annotated[bool, BoolType(nullable=False)]
Timestamp = typing.Annotated[datetime.datetime, TimestampType(nullable=False)]
Date = typing.Annotated[datetime.date, DateType(nullable=False)]


class _PxtType:
    """
    Base class for the Pixeltable type-hint family. Subclasses of this class are meant to be used as type hints, both
    in schema definitions and in UDF signatures. Whereas `ColumnType`s are instantiable and carry semantic information
    about the Pixeltable type system, `_PxtType` subclasses are purely for convenience: they are not instantiable and
    must be resolved to a `ColumnType` (by calling `ColumnType.from_python_type()`) in order to do anything meaningful
    with them.

    `_PxtType` subclasses can be specialized (as type hints) with type parameters; for example:
    `Image[(300, 300), 'RGB']`. The specialized forms resolve to `typing.Annotated` instances whose annotation is a
    `ColumnType`.
    """

    def __init__(self) -> None:
        raise TypeError(f'Type `{type(self)}` cannot be instantiated.')

    @classmethod
    def as_col_type(cls, nullable: bool) -> ColumnType:
        raise NotImplementedError()


class Json(_PxtType):
    def __class_getitem__(cls, item: Any) -> _AnnotatedAlias:
        """
        `item` (the type subscript) must be a `dict` representing a valid JSON Schema.
        """
        if not isinstance(item, dict):
            raise TypeError('Json type parameter must be a dict')

        # The JsonType initializer will validate the JSON Schema.
        return typing.Annotated[Any, JsonType(json_schema=item, nullable=False)]

    @classmethod
    def as_col_type(cls, nullable: bool) -> ColumnType:
        return JsonType(nullable=nullable)


class Array(np.ndarray, _PxtType):
    def __class_getitem__(cls, item: Any) -> _AnnotatedAlias:
        """
        `item` (the type subscript) must be a tuple with exactly two elements (in any order):
        - A tuple of `Optional[int]`s, specifying the shape of the array
        - A type, specifying the dtype of the array
        Example: Array[(3, None, 2), pxt.Float]
        """
        params = item if isinstance(item, tuple) else (item,)
        shape: Optional[tuple] = None
        dtype: Optional[ColumnType] = None
        if not any(isinstance(param, (type, _AnnotatedAlias)) for param in params):
            raise TypeError('Array type parameter must include a dtype.')
        for param in params:
            if isinstance(param, tuple):
                if not all(n is None or (isinstance(n, int) and n >= 1) for n in param):
                    raise TypeError(f'Invalid Array type parameter: {param}')
                if shape is not None:
                    raise TypeError(f'Duplicate Array type parameter: {param}')
                shape = param
            elif isinstance(param, (type, _AnnotatedAlias)):
                if dtype is not None:
                    raise TypeError(f'Duplicate Array type parameter: {param}')
                dtype = ColumnType.normalize_type(param, allow_builtin_types=False)
            else:
                raise TypeError(f'Invalid Array type parameter: {param}')
        return typing.Annotated[np.ndarray, ArrayType(shape=shape, dtype=dtype, nullable=False)]

    @classmethod
    def as_col_type(cls, nullable: bool) -> ColumnType:
        return ArrayType(nullable=nullable)


class Image(PIL.Image.Image, _PxtType):
    def __class_getitem__(cls, item: Any) -> _AnnotatedAlias:
        """
        `item` (the type subscript) must be one of the following, or a tuple containing either or both in any order:
        - A 2-tuple of `int`s, specifying the size of the image
        - A string, specifying the mode of the image
        Example: Image[(300, 300), 'RGB']
        """
        if isinstance(item, tuple) and all(n is None or isinstance(n, int) for n in item):
            # It's a tuple of the form (width, height)
            params = (item,)
        elif isinstance(item, tuple):
            # It's a compound tuple (multiple parameters)
            params = item
        else:
            # Not a tuple (single arg)
            params = (item,)
        size: Optional[tuple] = None
        mode: Optional[str] = None
        for param in params:
            if isinstance(param, tuple):
                if (
                    len(param) != 2
                    or not isinstance(param[0], (int, type(None)))
                    or not isinstance(param[1], (int, type(None)))
                ):
                    raise TypeError(f'Invalid Image type parameter: {param}')
                if size is not None:
                    raise TypeError(f'Duplicate Image type parameter: {param}')
                size = param
            elif isinstance(param, str):
                if param not in PIL.Image.MODES:
                    raise TypeError(f'Invalid Image type parameter: {param!r}')
                if mode is not None:
                    raise TypeError(f'Duplicate Image type parameter: {param!r}')
                mode = param
            else:
                raise TypeError(f'Invalid Image type parameter: {param}')
        return typing.Annotated[PIL.Image.Image, ImageType(size=size, mode=mode, nullable=False)]

    @classmethod
    def as_col_type(cls, nullable: bool) -> ColumnType:
        return ImageType(nullable=nullable)


class Video(str, _PxtType):
    @classmethod
    def as_col_type(cls, nullable: bool) -> ColumnType:
        return VideoType(nullable=nullable)


class Audio(str, _PxtType):
    @classmethod
    def as_col_type(cls, nullable: bool) -> ColumnType:
        return AudioType(nullable=nullable)


class Document(str, _PxtType):
    @classmethod
    def as_col_type(cls, nullable: bool) -> ColumnType:
        return DocumentType(nullable=nullable)


ALL_PIXELTABLE_TYPES = (String, Bool, Int, Float, Timestamp, Json, Array, Image, Video, Audio, Document)



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs) for `VideoType`.
"""

from typing import Any, Optional

import av
import numpy as np
import PIL.Image

import pixeltable as pxt
from pixeltable import env
from pixeltable.utils.code import local_public_names

_format_defaults = {  # format -> (codec, ext)
    'wav': ('pcm_s16le', 'wav'),
    'mp3': ('libmp3lame', 'mp3'),
    'flac': ('flac', 'flac'),
    # 'mp4': ('aac', 'm4a'),
}

# for mp4:
# - extract_audio() fails with
#   "Application provided invalid, non monotonically increasing dts to muxer in stream 0: 1146 >= 290"
# - chatgpt suggests this can be fixed in the following manner
#     for packet in container.demux(audio_stream):
#         packet.pts = None  # Reset the PTS and DTS to allow FFmpeg to set them automatically
#         packet.dts = None
#         for frame in packet.decode():
#             frame.pts = None
#             for packet in output_stream.encode(frame):
#                 output_container.mux(packet)
#
#     # Flush remaining packets
#     for packet in output_stream.encode():
#         output_container.mux(packet)


@pxt.uda(requires_order_by=True)
class make_video(pxt.Aggregator):
    """
    Aggregator that creates a video from a sequence of images.
    """

    container: Optional[av.container.OutputContainer]
    stream: Optional[av.video.stream.VideoStream]
    fps: int

    def __init__(self, fps: int = 25):
        """follows https://pyav.org/docs/develop/cookbook/numpy.html#generating-video"""
        self.container = None
        self.stream = None
        self.fps = fps

    def update(self, frame: PIL.Image.Image) -> None:
        if frame is None:
            return
        if self.container is None:
            self.out_file = env.Env.get().create_tmp_path('.mp4')
            self.container = av.open(str(self.out_file), mode='w')
            self.stream = self.container.add_stream('h264', rate=self.fps)
            self.stream.pix_fmt = 'yuv420p'
            self.stream.width = frame.width
            self.stream.height = frame.height

        av_frame = av.VideoFrame.from_ndarray(np.array(frame.convert('RGB')), format='rgb24')
        for packet in self.stream.encode(av_frame):
            self.container.mux(packet)

    def value(self) -> pxt.Video:
        for packet in self.stream.encode():
            self.container.mux(packet)
        self.container.close()
        return str(self.out_file)


@pxt.udf(is_method=True)
def extract_audio(
    video_path: pxt.Video, stream_idx: int = 0, format: str = 'wav', codec: Optional[str] = None
) -> pxt.Audio:
    """
    Extract an audio stream from a video.

    Args:
        stream_idx: Index of the audio stream to extract.
        format: The target audio format. (`'wav'`, `'mp3'`, `'flac'`).
        codec: The codec to use for the audio stream. If not provided, a default codec will be used.

    Returns:
        The extracted audio.

    Examples:
        Add a computed column to a table `tbl` that extracts audio from an existing column `video_col`:

        >>> tbl.add_computed_column(
        ...     extracted_audio=tbl.video_col.extract_audio(format='flac')
        ... )
    """
    if format not in _format_defaults:
        raise ValueError(f'extract_audio(): unsupported audio format: {format}')
    default_codec, ext = _format_defaults[format]

    with av.open(video_path) as container:
        if len(container.streams.audio) <= stream_idx:
            return None
        audio_stream = container.streams.audio[stream_idx]
        # create this in our tmp directory, so it'll get cleaned up if it's being generated as part of a query
        output_filename = str(env.Env.get().create_tmp_path(f'.{ext}'))

        with av.open(output_filename, 'w', format=format) as output_container:
            output_stream = output_container.add_stream(codec or default_codec)
            assert isinstance(output_stream, av.audio.stream.AudioStream)
            for packet in container.demux(audio_stream):
                for frame in packet.decode():
                    output_container.mux(output_stream.encode(frame))  # type: ignore[arg-type]

        return output_filename


@pxt.udf(is_method=True)
def get_metadata(video: pxt.Video) -> dict:
    """
    Gets various metadata associated with a video file and returns it as a dictionary.

    Args:
        video: The video to get metadata for.

    Returns:
        A `dict` such as the following:

            ```json
            {
                'bit_exact': False,
                'bit_rate': 967260,
                'size': 2234371,
                'metadata': {
                    'encoder': 'Lavf60.16.100',
                    'major_brand': 'isom',
                    'minor_version': '512',
                    'compatible_brands': 'isomiso2avc1mp41',
                },
                'streams': [
                    {
                        'type': 'video',
                        'width': 640,
                        'height': 360,
                        'frames': 462,
                        'time_base': 1.0 / 12800,
                        'duration': 236544,
                        'duration_seconds': 236544.0 / 12800,
                        'average_rate': 25.0,
                        'base_rate': 25.0,
                        'guessed_rate': 25.0,
                        'metadata': {
                            'language': 'und',
                            'handler_name': 'L-SMASH Video Handler',
                            'vendor_id': '[0][0][0][0]',
                            'encoder': 'Lavc60.31.102 libx264',
                        },
                        'codec_context': {'name': 'h264', 'codec_tag': 'avc1', 'profile': 'High', 'pix_fmt': 'yuv420p'},
                    }
                ],
            }
            ```

    Examples:
        Extract metadata for files in the `video_col` column of the table `tbl`:

        >>> tbl.select(tbl.video_col.get_metadata()).collect()
    """
    return _get_metadata(video)


def _get_metadata(path: str) -> dict:
    with av.open(path) as container:
        assert isinstance(container, av.container.InputContainer)
        streams_info = [__get_stream_metadata(stream) for stream in container.streams]
        result = {
            'bit_exact': getattr(container, 'bit_exact', False),
            'bit_rate': container.bit_rate,
            'size': container.size,
            'metadata': container.metadata,
            'streams': streams_info,
        }
    return result


def __get_stream_metadata(stream: av.stream.Stream) -> dict:
    if stream.type not in ('audio', 'video'):
        return {'type': stream.type}  # Currently unsupported

    codec_context = stream.codec_context
    codec_context_md: dict[str, Any] = {
        'name': codec_context.name,
        'codec_tag': codec_context.codec_tag.encode('unicode-escape').decode('utf-8'),
        'profile': codec_context.profile,
    }
    metadata = {
        'type': stream.type,
        'duration': stream.duration,
        'time_base': float(stream.time_base) if stream.time_base is not None else None,
        'duration_seconds': float(stream.duration * stream.time_base)
        if stream.duration is not None and stream.time_base is not None
        else None,
        'frames': stream.frames,
        'metadata': stream.metadata,
        'codec_context': codec_context_md,
    }

    if stream.type == 'audio':
        # Additional metadata for audio
        channels = getattr(stream.codec_context, 'channels', None)
        codec_context_md['channels'] = int(channels) if channels is not None else None
    else:
        assert stream.type == 'video'
        assert isinstance(stream, av.video.stream.VideoStream)
        # Additional metadata for video
        codec_context_md['pix_fmt'] = getattr(stream.codec_context, 'pix_fmt', None)
        metadata.update(
            **{
                'width': stream.width,
                'height': stream.height,
                'frames': stream.frames,
                'average_rate': float(stream.average_rate) if stream.average_rate is not None else None,
                'base_rate': float(stream.base_rate) if stream.base_rate is not None else None,
                'guessed_rate': float(stream.guessed_rate) if stream.guessed_rate is not None else None,
            }
        )

    return metadata


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import importlib
import inspect
from types import ModuleType
from typing import Optional

import pixeltable.exceptions as excs


def resolve_symbol(symbol_path: str) -> Optional[object]:
    path_elems = symbol_path.split('.')
    module: Optional[ModuleType] = None
    i = len(path_elems) - 1
    while i > 0 and module is None:
        try:
            module = importlib.import_module('.'.join(path_elems[:i]))
        except ModuleNotFoundError:
            i -= 1
    if i == 0:
        return None  # Not resolvable
    obj = module
    for el in path_elems[i:]:
        obj = getattr(obj, el)
    return obj


def validate_symbol_path(fn_path: str) -> None:
    path_elems = fn_path.split('.')
    fn_name = path_elems[-1]
    if any(el == '<locals>' for el in path_elems):
        raise excs.Error(
            f'{fn_name}(): nested functions are not supported. Move the function to the module level or into a class.'
        )
    if any(not el.isidentifier() for el in path_elems):
        raise excs.Error(
            f'{fn_name}(): cannot resolve symbol path {fn_path}. Move the function to the module level or into a class.'
        )


def get_caller_module_path() -> str:
    """Return the module path of our caller's caller"""
    stack = inspect.stack()
    try:
        caller_frame = stack[2].frame
        module_path = caller_frame.f_globals['__name__']
    finally:
        # remove references to stack frames to avoid reference cycles
        del stack
    return module_path



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various endpoints from the Mistral AI API. In order to use them, you must
first `pip install mistralai` and configure your Mistral AI credentials, as described in
the [Working with Mistral AI](https://pixeltable.readme.io/docs/working-with-mistralai) tutorial.
"""

from typing import TYPE_CHECKING, Any, Optional

import numpy as np

import pixeltable as pxt
import pixeltable.type_system as ts
from pixeltable.env import Env, register_client
from pixeltable.func.signature import Batch
from pixeltable.utils.code import local_public_names

if TYPE_CHECKING:
    import mistralai


@register_client('mistral')
def _(api_key: str) -> 'mistralai.Mistral':
    import mistralai

    return mistralai.Mistral(api_key=api_key)


def _mistralai_client() -> 'mistralai.Mistral':
    return Env.get().get_client('mistral')


@pxt.udf(resource_pool='request-rate:mistral')
async def chat_completions(
    messages: list[dict[str, str]], *, model: str, model_kwargs: Optional[dict[str, Any]] = None
) -> dict:
    """
    Chat Completion API.

    Equivalent to the Mistral AI `chat/completions` API endpoint.
    For additional details, see: <https://docs.mistral.ai/api/#tag/chat>

    Request throttling:
    Applies the rate limit set in the config (section `mistral`, key `rate_limit`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install mistralai`

    Args:
        messages: The prompt(s) to generate completions for.
        model: ID of the model to use. (See overview here: <https://docs.mistral.ai/getting-started/models/>)
        model_kwargs: Additional keyword args for the Mistral `chat/completions` API.
            For details on the available parameters, see: <https://docs.mistral.ai/api/#tag/chat>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `mistral-latest-small`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> messages = [{'role': 'user', 'content': tbl.prompt}]
        ... tbl.add_computed_column(response=completions(messages, model='mistral-latest-small'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    Env.get().require_package('mistralai')
    result = await _mistralai_client().chat.complete_async(
        messages=messages,  # type: ignore[arg-type]
        model=model,
        **model_kwargs,
    )
    return result.dict()


@pxt.udf(resource_pool='request-rate:mistral')
async def fim_completions(prompt: str, *, model: str, model_kwargs: Optional[dict[str, Any]] = None) -> dict:
    """
    Fill-in-the-middle Completion API.

    Equivalent to the Mistral AI `fim/completions` API endpoint.
    For additional details, see: <https://docs.mistral.ai/api/#tag/fim>

    Request throttling:
    Applies the rate limit set in the config (section `mistral`, key `rate_limit`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install mistralai`

    Args:
        prompt: The text/code to complete.
        model: ID of the model to use. (See overview here: <https://docs.mistral.ai/getting-started/models/>)
        model_kwargs: Additional keyword args for the Mistral `fim/completions` API.
            For details on the available parameters, see: <https://docs.mistral.ai/api/#tag/fim>

    For details on the other parameters, see: <https://docs.mistral.ai/api/#tag/fim>

    Returns:
        A dictionary containing the response and other metadata.

    Examples:
        Add a computed column that applies the model `codestral-latest`
        to an existing Pixeltable column `tbl.prompt` of the table `tbl`:

        >>> tbl.add_computed_column(response=completions(tbl.prompt, model='codestral-latest'))
    """
    if model_kwargs is None:
        model_kwargs = {}

    Env.get().require_package('mistralai')
    result = await _mistralai_client().fim.complete_async(prompt=prompt, model=model, **model_kwargs)
    return result.dict()


_embedding_dimensions_cache: dict[str, int] = {'mistral-embed': 1024}


@pxt.udf(batch_size=16, resource_pool='request-rate:mistral')
async def embeddings(input: Batch[str], *, model: str) -> Batch[pxt.Array[(None,), pxt.Float]]:  # noqa: RUF029
    """
    Embeddings API.

    Equivalent to the Mistral AI `embeddings` API endpoint.
    For additional details, see: <https://docs.mistral.ai/api/#tag/embeddings>

    Request throttling:
    Applies the rate limit set in the config (section `mistral`, key `rate_limit`). If no rate
    limit is configured, uses a default of 600 RPM.

    __Requirements:__

    - `pip install mistralai`

    Args:
        input: Text to embed.
        model: ID of the model to use. (See overview here: <https://docs.mistral.ai/getting-started/models/>)

    Returns:
        An array representing the application of the given embedding to `input`.
    """
    Env.get().require_package('mistralai')
    result = _mistralai_client().embeddings.create(inputs=input, model=model)
    return [np.array(data.embedding, dtype=np.float64) for data in result.data]


@embeddings.conditional_return_type
def _(model: str) -> ts.ArrayType:
    dimensions = _embedding_dimensions_cache.get(model)  # `None` if unknown model
    return ts.ArrayType((dimensions,), dtype=ts.FloatType())


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



from __future__ import annotations

import logging
from typing import Iterator, Optional

from pixeltable import catalog, exprs
from pixeltable.utils.media_store import MediaStore

_logger = logging.getLogger('pixeltable')


class DataRowBatch:
    """Set of DataRows, indexed by rowid.

    Contains the metadata needed to initialize DataRows.
    """

    tbl: Optional[catalog.TableVersionHandle]
    row_builder: exprs.RowBuilder
    img_slot_idxs: list[int]
    media_slot_idxs: list[int]  # non-image media slots
    array_slot_idxs: list[int]
    rows: list[exprs.DataRow]

    def __init__(
        self,
        tbl: Optional[catalog.TableVersionHandle],
        row_builder: exprs.RowBuilder,
        num_rows: Optional[int] = None,
        rows: Optional[list[exprs.DataRow]] = None,
    ):
        """
        Requires either num_rows or rows to be specified, but not both.
        """
        assert num_rows is None or rows is None
        self.tbl = tbl
        self.row_builder = row_builder
        self.img_slot_idxs = [e.slot_idx for e in row_builder.unique_exprs if e.col_type.is_image_type()]
        # non-image media slots
        self.media_slot_idxs = [
            e.slot_idx
            for e in row_builder.unique_exprs
            if e.col_type.is_media_type() and not e.col_type.is_image_type()
        ]
        self.array_slot_idxs = [e.slot_idx for e in row_builder.unique_exprs if e.col_type.is_array_type()]
        if rows is not None:
            self.rows = rows
        else:
            if num_rows is None:
                num_rows = 0
            self.rows = [
                exprs.DataRow(
                    row_builder.num_materialized, self.img_slot_idxs, self.media_slot_idxs, self.array_slot_idxs
                )
                for _ in range(num_rows)
            ]

    def add_row(self, row: Optional[exprs.DataRow] = None) -> exprs.DataRow:
        if row is None:
            row = exprs.DataRow(
                self.row_builder.num_materialized, self.img_slot_idxs, self.media_slot_idxs, self.array_slot_idxs
            )
        self.rows.append(row)
        return row

    def pop_row(self) -> exprs.DataRow:
        return self.rows.pop()

    def __len__(self) -> int:
        return len(self.rows)

    def __getitem__(self, index: int) -> exprs.DataRow:
        return self.rows[index]

    def flush_imgs(
        self,
        idx_range: Optional[slice] = None,
        stored_img_info: Optional[list[exprs.ColumnSlotIdx]] = None,
        flushed_slot_idxs: Optional[list[int]] = None,
    ) -> None:
        """Flushes images in the given range of rows."""
        assert self.tbl is not None
        if stored_img_info is None:
            stored_img_info = []
        if flushed_slot_idxs is None:
            flushed_slot_idxs = []
        if len(stored_img_info) == 0 and len(flushed_slot_idxs) == 0:
            return
        if idx_range is None:
            idx_range = slice(0, len(self.rows))
        for row in self.rows[idx_range]:
            for info in stored_img_info:
                col = info.col
                assert col.tbl.id == self.tbl.id
                filepath = str(MediaStore.prepare_media_path(col.tbl.id, col.id, col.tbl.version))
                row.flush_img(info.slot_idx, filepath)
            for slot_idx in flushed_slot_idxs:
                row.flush_img(slot_idx)

    def __iter__(self) -> Iterator[exprs.DataRow]:
        return iter(self.rows)



import dataclasses
import importlib
import logging
import os
import pkgutil
from typing import Callable

import sqlalchemy as sql
from sqlalchemy import orm

import pixeltable as pxt
import pixeltable.exceptions as excs
from pixeltable.utils.console_output import ConsoleLogger

from .schema import SystemInfo, SystemInfoMd

_console_logger = ConsoleLogger(logging.getLogger('pixeltable'))
_logger = logging.getLogger('pixeltable')

# current version of the metadata; this is incremented whenever the metadata schema changes
VERSION = 40


def create_system_info(engine: sql.engine.Engine) -> None:
    """Create the system metadata record"""
    system_md = SystemInfoMd(schema_version=VERSION)
    record = SystemInfo(md=dataclasses.asdict(system_md))
    with orm.Session(engine, future=True) as session:
        # Write system metadata only once for idempotency
        if session.query(SystemInfo).count() == 0:
            session.add(record)
            session.flush()
            session.commit()


# conversion functions for upgrading the metadata schema from one version to the following
# key: old schema version
converter_cbs: dict[int, Callable[[sql.engine.Engine], None]] = {}


def register_converter(version: int) -> Callable[[Callable[[sql.engine.Engine], None]], None]:
    def decorator(fn: Callable[[sql.engine.Engine], None]) -> None:
        assert version not in converter_cbs
        converter_cbs[version] = fn

    return decorator


# load all converter modules
for _, modname, _ in pkgutil.iter_modules([os.path.dirname(__file__) + '/converters']):
    importlib.import_module('pixeltable.metadata.converters.' + modname)


def upgrade_md(engine: sql.engine.Engine) -> None:
    """Upgrade the metadata schema to the current version"""
    with orm.Session(engine) as session:
        system_info = session.query(SystemInfo).one().md
        md_version = system_info['schema_version']
        assert isinstance(md_version, int)
        _logger.info(f'Current database version: {md_version}, installed version: {VERSION}')
        if md_version > VERSION:
            raise excs.Error(
                'This Pixeltable database was created with a newer Pixeltable version '
                f'than the one currently installed ({pxt.__version__}).\n'
                'Please update to the latest Pixeltable version by running: pip install --upgrade pixeltable'
            )
        if md_version == VERSION:
            return
        while md_version < VERSION:
            if md_version not in converter_cbs:
                raise RuntimeError(f'No metadata converter for version {md_version}')
            # We can't use the console logger in Env, because Env might not have been initialized yet.
            _console_logger.info(f'Converting metadata from version {md_version} to {md_version + 1}')
            converter_cbs[md_version](engine)
            md_version += 1
        # update system info
        conn = session.connection()
        system_info_md = SystemInfoMd(schema_version=VERSION)
        conn.execute(SystemInfo.__table__.update().values(md=dataclasses.asdict(system_info_md)))
        session.commit()



from __future__ import annotations

import copy
import dataclasses
import importlib
import logging
import time
import uuid
from typing import TYPE_CHECKING, Any, Iterable, Iterator, Literal, Optional, Tuple
from uuid import UUID

import jsonschema.exceptions
import sqlalchemy as sql

import pixeltable as pxt
import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from pixeltable import exprs, index
from pixeltable.env import Env
from pixeltable.iterators import ComponentIterator
from pixeltable.metadata import schema
from pixeltable.utils.exception_handler import run_cleanup_on_exception
from pixeltable.utils.filecache import FileCache
from pixeltable.utils.media_store import MediaStore

from .tbl_ops import TableOp

if TYPE_CHECKING:
    from pixeltable.plan import SampleClause

from ..func.globals import resolve_symbol
from .column import Column
from .globals import _POS_COLUMN_NAME, _ROWID_COLUMN_NAME, MediaValidation, is_valid_identifier
from .update_status import RowCountStats, UpdateStatus

if TYPE_CHECKING:
    from pixeltable import exec, store

    from .table_version_handle import TableVersionHandle

_logger = logging.getLogger('pixeltable')


@dataclasses.dataclass(frozen=True)
class TableVersionMd:
    """
    Complete set of md records for a specific TableVersion instance.

    TODO: subsume schema.FullTableMd
    """

    tbl_md: schema.TableMd
    version_md: schema.TableVersionMd
    schema_version_md: schema.TableSchemaVersionMd


class TableVersion:
    """
    TableVersion represents a particular version of a table/view along with its physical representation:
    - the physical representation is a store table with indices
    - the version can be mutable or a snapshot
    - tables and their recursive views form a tree, and a mutable TableVersion also records its own
      mutable views in order to propagate updates
    - each view TableVersion records its base:
      * the base is correct only for mutable views (snapshot versions form a DAG, not a tree)
      * the base is useful for getting access to the StoreTable and the base id
      * TODO: create a separate hierarchy of objects that records the version-independent tree of tables/views, and
        have TableVersions reference those
    - mutable TableVersions record their TableVersionPath, which is needed for expr evaluation in updates

    Instances of TableVersion should not be stored as member variables (ie, used across transaction boundaries).
    Use a TableVersionHandle instead.

    Only TableVersion and Catalog interact directly with stored metadata. Everything else needs to go through these
    two classes.
    """

    id: UUID

    # record metadata stored in catalog
    _tbl_md: schema.TableMd
    _version_md: schema.TableVersionMd
    _schema_version_md: schema.TableSchemaVersionMd

    effective_version: Optional[int]
    path: Optional[pxt.catalog.TableVersionPath]  # only set for live tables; needed to resolve computed cols
    base: Optional[TableVersionHandle]  # only set for views
    predicate: Optional[exprs.Expr]
    sample_clause: Optional['SampleClause']

    iterator_cls: Optional[type[ComponentIterator]]
    iterator_args: Optional[exprs.InlineDict]
    num_iterator_cols: int

    # target for data operation propagation (only set for non-snapshots, and only records non-snapshot views)
    mutable_views: frozenset[TableVersionHandle]

    # contains complete history of columns, incl dropped ones
    cols: list[Column]
    # contains only user-facing (named) columns visible in this version
    cols_by_name: dict[str, Column]
    # contains only columns visible in this version, both system and user
    cols_by_id: dict[int, Column]
    # contains only actively maintained indices
    idxs_by_name: dict[str, TableVersion.IndexInfo]

    external_stores: dict[str, pxt.io.ExternalStore]
    store_tbl: Optional['store.StoreBase']

    is_initialized: bool  # True if init() has been called

    # used by Catalog to invalidate cached instances at the end of a transaction;
    # True if this instance reflects the state of stored metadata in the context of this transaction and
    # it is the instance cached in Catalog
    is_validated: bool

    @dataclasses.dataclass
    class IndexInfo:
        id: int
        name: str
        idx: index.IndexBase
        col: Column
        val_col: Column
        undo_col: Column

    def __init__(
        self,
        id: UUID,
        tbl_md: schema.TableMd,
        version_md: schema.TableVersionMd,
        effective_version: Optional[int],
        schema_version_md: schema.TableSchemaVersionMd,
        mutable_views: list[TableVersionHandle],
        base_path: Optional[pxt.catalog.TableVersionPath] = None,
        base: Optional[TableVersionHandle] = None,
    ):
        self.is_validated = True  # a freshly constructed instance is always valid
        self.is_initialized = False
        self.id = id
        self._tbl_md = copy.deepcopy(tbl_md)
        self._version_md = copy.deepcopy(version_md)
        self._schema_version_md = copy.deepcopy(schema_version_md)
        self.effective_version = effective_version
        assert not (self.is_view and base is None)
        self.base = base
        self.store_tbl = None

        # mutable tables need their TableVersionPath for expr eval during updates
        from .table_version_handle import TableVersionHandle
        from .table_version_path import TableVersionPath

        if self.is_snapshot:
            self.path = None
        else:
            self_handle = TableVersionHandle(id, self.effective_version)
            if self.is_view:
                assert base_path is not None
            self.path = TableVersionPath(self_handle, base=base_path)

        # view-specific initialization
        from pixeltable import exprs
        from pixeltable.plan import SampleClause

        predicate_dict = None if self.view_md is None or self.view_md.predicate is None else self.view_md.predicate
        self.predicate = exprs.Expr.from_dict(predicate_dict) if predicate_dict is not None else None
        sample_dict = None if self.view_md is None or self.view_md.sample_clause is None else self.view_md.sample_clause
        self.sample_clause = SampleClause.from_dict(sample_dict) if sample_dict is not None else None

        # component view-specific initialization
        self.iterator_cls = None
        self.iterator_args = None
        self.num_iterator_cols = 0
        if self.view_md is not None and self.view_md.iterator_class_fqn is not None:
            module_name, class_name = tbl_md.view_md.iterator_class_fqn.rsplit('.', 1)
            module = importlib.import_module(module_name)
            self.iterator_cls = getattr(module, class_name)
            self.iterator_args = exprs.InlineDict.from_dict(tbl_md.view_md.iterator_args)
            output_schema, _ = self.iterator_cls.output_schema(**self.iterator_args.to_kwargs())
            self.num_iterator_cols = len(output_schema)
            assert tbl_md.view_md.iterator_args is not None

        self.mutable_views = frozenset(mutable_views)
        assert self.is_mutable or len(self.mutable_views) == 0

        self.cols = []
        self.cols_by_name = {}
        self.cols_by_id = {}
        self.idxs_by_name = {}
        self.external_stores = {}

    def __hash__(self) -> int:
        return hash(self.id)

    def create_snapshot_copy(self) -> TableVersion:
        """Create a snapshot copy of this TableVersion"""
        assert not self.is_snapshot
        base = self.path.base.tbl_version if self.is_view else None
        return TableVersion(
            self.id, self.tbl_md, self.version_md, self.version, self.schema_version_md, mutable_views=[], base=base
        )

    @property
    def versioned_name(self) -> str:
        if self.effective_version is None:
            return self.name
        else:
            return f'{self.name}:{self.effective_version}'

    @property
    def handle(self) -> 'TableVersionHandle':
        from .table_version_handle import TableVersionHandle

        return TableVersionHandle(self.id, self.effective_version, self)

    @classmethod
    def create_initial_md(
        cls,
        name: str,
        cols: list[Column],
        num_retained_versions: int,
        comment: str,
        media_validation: MediaValidation,
        view_md: Optional[schema.ViewMd] = None,
    ) -> TableVersionMd:
        user = Env.get().user

        # assign ids
        cols_by_name: dict[str, Column] = {}
        for pos, col in enumerate(cols):
            col.id = pos
            col.schema_version_add = 0
            cols_by_name[col.name] = col
            if col.is_computed:
                col.check_value_expr()

        timestamp = time.time()
        column_md = cls._create_column_md(cols)
        tbl_id = uuid.uuid4()
        tbl_id_str = str(tbl_id)
        tbl_md = schema.TableMd(
            tbl_id=tbl_id_str,
            name=name,
            user=user,
            is_replica=False,
            current_version=0,
            current_schema_version=0,
            next_col_id=len(cols),
            next_idx_id=0,
            next_row_id=0,
            view_sn=0,
            column_md=column_md,
            index_md={},
            external_stores=[],
            view_md=view_md,
            additional_md={},
        )

        table_version_md = schema.TableVersionMd(
            tbl_id=tbl_id_str, created_at=timestamp, version=0, schema_version=0, additional_md={}
        )

        schema_col_md: dict[int, schema.SchemaColumn] = {}
        for pos, col in enumerate(cols):
            md = schema.SchemaColumn(
                pos=pos,
                name=col.name,
                media_validation=col._media_validation.name.lower() if col._media_validation is not None else None,
            )
            schema_col_md[col.id] = md

        schema_version_md = schema.TableSchemaVersionMd(
            tbl_id=tbl_id_str,
            schema_version=0,
            preceding_schema_version=None,
            columns=schema_col_md,
            num_retained_versions=num_retained_versions,
            comment=comment,
            media_validation=media_validation.name.lower(),
            additional_md={},
        )
        return TableVersionMd(tbl_md, table_version_md, schema_version_md)

    @classmethod
    def create(
        cls,
        dir_id: UUID,
        name: str,
        cols: list[Column],
        num_retained_versions: int,
        comment: str,
        media_validation: MediaValidation,
    ) -> tuple[UUID, Optional[TableVersion]]:
        user = Env.get().user

        # assign ids
        cols_by_name: dict[str, Column] = {}
        for pos, col in enumerate(cols):
            col.id = pos
            col.schema_version_add = 0
            cols_by_name[col.name] = col
            if col.is_computed:
                col.check_value_expr()

        timestamp = time.time()
        # create schema.Table
        # Column.dependent_cols for existing cols is wrong at this point, but init() will set it correctly
        column_md = cls._create_column_md(cols)
        tbl_id = uuid.uuid4()
        tbl_id_str = str(tbl_id)
        table_md = schema.TableMd(
            tbl_id=tbl_id_str,
            name=name,
            user=user,
            is_replica=False,
            current_version=0,
            current_schema_version=0,
            next_col_id=len(cols),
            next_idx_id=0,
            next_row_id=0,
            view_sn=0,
            column_md=column_md,
            index_md={},
            external_stores=[],
            view_md=None,
            additional_md={},
        )

        # create schema.TableVersion of the initial version
        table_version_md = schema.TableVersionMd(
            tbl_id=tbl_id_str,
            created_at=timestamp,
            version=0,
            schema_version=0,
            user=user,
            update_status=None,
            additional_md={},
        )

        # create schema.TableSchemaVersion
        schema_col_md: dict[int, schema.SchemaColumn] = {}
        for pos, col in enumerate(cols):
            md = schema.SchemaColumn(
                pos=pos,
                name=col.name,
                media_validation=col._media_validation.name.lower() if col._media_validation is not None else None,
            )
            schema_col_md[col.id] = md

        schema_version_md = schema.TableSchemaVersionMd(
            tbl_id=tbl_id_str,
            schema_version=0,
            preceding_schema_version=None,
            columns=schema_col_md,
            num_retained_versions=num_retained_versions,
            comment=comment,
            media_validation=media_validation.name.lower(),
            additional_md={},
        )

        cat = pxt.catalog.Catalog.get()

        tbl_version = cls(tbl_id, table_md, table_version_md, None, schema_version_md, [])
        # TODO: break this up, so that Catalog.create_table() registers tbl_version
        cat._tbl_versions[tbl_id, None] = tbl_version
        tbl_version.init()
        tbl_version.store_tbl.create()
        # add default indices, after creating the store table
        for col in tbl_version.cols_by_name.values():
            status = tbl_version._add_default_index(col)
            assert status is None or status.num_excs == 0

        cat.store_tbl_md(
            tbl_id=tbl_id,
            dir_id=dir_id,
            tbl_md=tbl_version.tbl_md,
            version_md=table_version_md,
            schema_version_md=schema_version_md,
        )
        return tbl_id, tbl_version

    def exec_op(self, op: TableOp) -> None:
        if op.create_store_table_op is not None:
            # don't use Catalog.begin_xact() here, to avoid accidental recursive calls to exec_op()
            with Env.get().begin_xact():
                self.store_tbl.create()

        elif op.load_view_op is not None:
            from pixeltable.catalog import Catalog
            from pixeltable.plan import Planner

            from .table_version_path import TableVersionPath

            # clear out any remaining media files from an aborted previous attempt
            MediaStore.delete(self.id)
            view_path = TableVersionPath.from_dict(op.load_view_op.view_path)
            plan, _ = Planner.create_view_load_plan(view_path)
            _, row_counts = self.store_tbl.insert_rows(plan, v_min=self.version)
            status = UpdateStatus(row_count_stats=row_counts)
            Catalog.get().store_update_status(self.id, self.version, status)
            _logger.debug(f'Loaded view {self.name} with {row_counts.num_rows} rows')

    @classmethod
    def create_replica(cls, md: schema.FullTableMd) -> TableVersion:
        assert Env.get().in_xact
        tbl_id = UUID(md.tbl_md.tbl_id)
        _logger.info(f'Creating replica table version {tbl_id}:{md.version_md.version}.')
        view_md = md.tbl_md.view_md
        base_path = pxt.catalog.TableVersionPath.from_md(view_md.base_versions) if view_md is not None else None
        base = base_path.tbl_version if base_path is not None else None
        tbl_version = cls(
            tbl_id,
            md.tbl_md,
            md.version_md,
            md.version_md.version,
            md.schema_version_md,
            [],
            base_path=base_path,
            base=base,
        )
        cat = pxt.catalog.Catalog.get()
        # We're creating a new TableVersion replica, so we should never have seen this particular
        # TableVersion instance before.
        assert tbl_version.effective_version is not None
        assert (tbl_version.id, tbl_version.effective_version) not in cat._tbl_versions
        cat._tbl_versions[tbl_version.id, tbl_version.effective_version] = tbl_version
        tbl_version.init()
        tbl_version.store_tbl.create()
        tbl_version.store_tbl.ensure_columns_exist(col for col in tbl_version.cols if col.is_stored)
        return tbl_version

    def drop(self) -> None:
        # if self.is_view and self.is_mutable:
        #     # update mutable_views
        #     # TODO: invalidate base to force reload
        #     from .table_version_handle import TableVersionHandle
        #
        #     assert self.base is not None
        #     if self.base.get().is_mutable:
        #         self.base.get().mutable_views.remove(TableVersionHandle.create(self))

        MediaStore.delete(self.id)
        FileCache.get().clear(tbl_id=self.id)
        self.store_tbl.drop()

    def init(self) -> None:
        """
        Initialize schema-related in-memory metadata separately, now that this TableVersion instance is visible
        in Catalog.
        """
        from .catalog import Catalog

        cat = Catalog.get()
        assert (self.id, self.effective_version) in cat._tbl_versions
        self._init_schema()
        if self.is_mutable:
            cat.record_column_dependencies(self)
        # init external stores; this needs to happen after the schema is created
        self._init_external_stores()
        self.is_initialized = True

    def _init_schema(self) -> None:
        # create columns first, so the indices can reference them
        self._init_cols()
        if not self.is_snapshot:
            self._init_idxs()
        # create the sa schema only after creating the columns and indices
        self._init_sa_schema()

        # created value_exprs after everything else has been initialized
        for col in self.cols_by_id.values():
            col.init_value_expr()

    def _init_cols(self) -> None:
        """Initialize self.cols with the columns visible in our effective version"""
        self.cols = []
        self.cols_by_name = {}
        self.cols_by_id = {}
        # Sort columns in column_md by the position specified in col_md.id to guarantee that all references
        # point backward.
        sorted_column_md = sorted(self.tbl_md.column_md.values(), key=lambda item: item.id)
        for col_md in sorted_column_md:
            schema_col_md = self.schema_version_md.columns.get(col_md.id)
            col_name = schema_col_md.name if schema_col_md is not None else None
            media_val = (
                MediaValidation[schema_col_md.media_validation.upper()]
                if schema_col_md is not None and schema_col_md.media_validation is not None
                else None
            )
            col = Column(
                col_id=col_md.id,
                name=col_name,
                col_type=ts.ColumnType.from_dict(col_md.col_type),
                is_pk=col_md.is_pk,
                stored=col_md.stored,
                media_validation=media_val,
                schema_version_add=col_md.schema_version_add,
                schema_version_drop=col_md.schema_version_drop,
                value_expr_dict=col_md.value_expr,
                tbl=self,
            )
            col.tbl = self
            self.cols.append(col)

            # populate the lookup structures before Expr.from_dict()
            if col_md.schema_version_add > self.schema_version:
                # column was added after this version
                continue
            if col_md.schema_version_drop is not None and col_md.schema_version_drop <= self.schema_version:
                # column was dropped
                continue
            if col.name is not None:
                self.cols_by_name[col.name] = col
            self.cols_by_id[col.id] = col

            # # make sure to traverse columns ordered by position = order in which cols were created;
            # # this guarantees that references always point backwards
            # if not self.is_snapshot and col_md.value_expr is not None:
            #     self._record_refd_columns(col)

    def _init_idxs(self) -> None:
        # self.idx_md = tbl_md.index_md
        self.idxs_by_name = {}
        import pixeltable.index as index_module

        for md in self.tbl_md.index_md.values():
            if md.schema_version_add > self.schema_version or (
                md.schema_version_drop is not None and md.schema_version_drop <= self.schema_version
            ):
                # index not visible in this schema version
                continue

            # instantiate index object
            cls_name = md.class_fqn.rsplit('.', 1)[-1]
            cls = getattr(index_module, cls_name)
            idx_col: Column
            if md.indexed_col_tbl_id == str(self.id):
                # this is a reference to one of our columns: avoid TVP.get_column_by_id() here, because we're not fully
                # initialized yet
                idx_col = self.cols_by_id[md.indexed_col_id]
            else:
                assert self.path.base is not None
                idx_col = self.path.base.get_column_by_id(UUID(md.indexed_col_tbl_id), md.indexed_col_id)
            idx = cls.from_dict(idx_col, md.init_args)

            # fix up the sa column type of the index value and undo columns
            val_col = self.cols_by_id[md.index_val_col_id]
            val_col.sa_col_type = idx.index_sa_type()
            val_col._stores_cellmd = False
            undo_col = self.cols_by_id[md.index_val_undo_col_id]
            undo_col.sa_col_type = idx.index_sa_type()
            undo_col._stores_cellmd = False
            idx_info = self.IndexInfo(id=md.id, name=md.name, idx=idx, col=idx_col, val_col=val_col, undo_col=undo_col)
            self.idxs_by_name[md.name] = idx_info

    def _init_sa_schema(self) -> None:
        # create the sqlalchemy schema; do this after instantiating columns, in order to determine whether they
        # need to record errors
        from pixeltable.store import StoreComponentView, StoreTable, StoreView

        if self.is_component_view:
            self.store_tbl = StoreComponentView(self)
        elif self.is_view:
            self.store_tbl = StoreView(self)
        else:
            self.store_tbl = StoreTable(self)

    def _write_md(self, new_version: bool, new_schema_version: bool) -> None:
        from pixeltable.catalog import Catalog

        Catalog.get().store_tbl_md(
            self.id,
            None,
            self._tbl_md,
            self._version_md if new_version else None,
            self._schema_version_md if new_schema_version else None,
        )

    def _store_idx_name(self, idx_id: int) -> str:
        """Return name of index in the store, which needs to be globally unique"""
        return f'idx_{self.id.hex}_{idx_id}'

    def add_index(self, col: Column, idx_name: Optional[str], idx: index.IndexBase) -> UpdateStatus:
        # we're creating a new schema version
        self.version += 1
        self.created_at = time.time()
        self.schema_version = self.version
        status = self._add_index(col, idx_name, idx)
        self._write_md(new_version=True, new_schema_version=True)
        _logger.info(f'Added index {idx_name} on column {col.name} to table {self.name}')
        return status

    def _is_btree_indexable(self, col: Column) -> bool:
        if not col.stored:
            # if the column is intentionally not stored, we want to avoid the overhead of an index
            return False
        # Skip index for stored media columns produced by an iterator
        if col.col_type.is_media_type() and self.is_iterator_column(col):
            return False
        if not col.col_type.is_scalar_type() and not (col.col_type.is_media_type() and not col.is_computed):
            # wrong type for a B-tree
            return False
        if col.col_type.is_bool_type():  # noqa : SIM103 Supress `Return the negated condition directly` check
            # B-trees on bools aren't useful
            return False
        return True

    def _add_default_index(self, col: Column) -> Optional[UpdateStatus]:
        """Add a B-tree index on this column if it has a compatible type"""
        if not self._is_btree_indexable(col):
            return None
        status = self._add_index(col, idx_name=None, idx=index.BtreeIndex(col))
        return status

    def _create_index_columns(self, idx: index.IndexBase) -> Tuple[Column, Column]:
        """Create value and undo columns for the given index.
        Args:
            idx:  index for which columns will be created.
        Returns:
            A tuple containing the value column and the undo column.
        """
        assert not self.is_snapshot
        # add the index value and undo columns (which need to be nullable)
        val_col = Column(
            col_id=self.next_col_id,
            name=None,
            computed_with=idx.index_value_expr(),
            sa_col_type=idx.index_sa_type(),
            stored=True,
            schema_version_add=self.schema_version,
            schema_version_drop=None,
            stores_cellmd=idx.records_value_errors(),
        )
        val_col.tbl = self
        val_col.col_type = val_col.col_type.copy(nullable=True)
        self.next_col_id += 1

        undo_col = Column(
            col_id=self.next_col_id,
            name=None,
            col_type=val_col.col_type,
            sa_col_type=val_col.sa_col_type,
            stored=True,
            schema_version_add=self.schema_version,
            schema_version_drop=None,
            stores_cellmd=False,
        )
        undo_col.tbl = self
        undo_col.col_type = undo_col.col_type.copy(nullable=True)
        self.next_col_id += 1
        return val_col, undo_col

    def _create_index(
        self, col: Column, val_col: Column, undo_col: Column, idx_name: Optional[str], idx: index.IndexBase
    ) -> None:
        """Create the given index along with index md"""
        idx_id = self.next_idx_id
        self.next_idx_id += 1
        if idx_name is None:
            idx_name = f'idx{idx_id}'
        else:
            assert is_valid_identifier(idx_name)
            assert idx_name not in [i.name for i in self._tbl_md.index_md.values()]
        # create and register the index metadata
        idx_cls = type(idx)
        idx_md = schema.IndexMd(
            id=idx_id,
            name=idx_name,
            indexed_col_id=col.id,
            indexed_col_tbl_id=str(col.tbl.id),
            index_val_col_id=val_col.id,
            index_val_undo_col_id=undo_col.id,
            schema_version_add=self.schema_version,
            schema_version_drop=None,
            class_fqn=idx_cls.__module__ + '.' + idx_cls.__name__,
            init_args=idx.as_dict(),
        )
        idx_info = self.IndexInfo(id=idx_id, name=idx_name, idx=idx, col=col, val_col=val_col, undo_col=undo_col)
        self._tbl_md.index_md[idx_id] = idx_md
        self.idxs_by_name[idx_name] = idx_info
        try:
            idx.create_index(self._store_idx_name(idx_id), val_col)
        finally:

            def cleanup_index() -> None:
                """Delete the newly added in-memory index structure"""
                del self.idxs_by_name[idx_name]
                del self._tbl_md.index_md[idx_id]
                self.next_idx_id = idx_id

            # Run cleanup only if there has been an exception; otherwise, skip cleanup.
            run_cleanup_on_exception(cleanup_index)

    def _add_index(self, col: Column, idx_name: Optional[str], idx: index.IndexBase) -> UpdateStatus:
        val_col, undo_vol = self._create_index_columns(idx)
        # add the columns and update the metadata
        # TODO support on_error='abort' for indices; it's tricky because of the way metadata changes are entangled
        # with the database operations
        status = self._add_columns([val_col, undo_vol], print_stats=False, on_error='ignore')
        # now create the index structure
        self._create_index(col, val_col, undo_vol, idx_name, idx)
        return status

    def drop_index(self, idx_id: int) -> None:
        assert self.is_mutable
        assert idx_id in self._tbl_md.index_md

        # we're creating a new schema version
        self.version += 1
        self.created_at = time.time()
        self.schema_version = self.version
        idx_md = self._tbl_md.index_md[idx_id]
        idx_md.schema_version_drop = self.schema_version
        assert idx_md.name in self.idxs_by_name
        idx_info = self.idxs_by_name[idx_md.name]
        # remove this index entry from the active indexes (in memory)
        # and the index metadata (in persistent table metadata)
        # TODO: this is wrong, it breaks revert()
        del self.idxs_by_name[idx_md.name]
        del self._tbl_md.index_md[idx_id]

        self._drop_columns([idx_info.val_col, idx_info.undo_col])
        self._write_md(new_version=True, new_schema_version=True)
        _logger.info(f'Dropped index {idx_md.name} on table {self.name}')

    def add_columns(
        self, cols: Iterable[Column], print_stats: bool, on_error: Literal['abort', 'ignore']
    ) -> UpdateStatus:
        """Adds columns to the table."""
        assert self.is_mutable
        assert all(is_valid_identifier(col.name) for col in cols if col.name is not None)
        assert all(col.stored is not None for col in cols)
        assert all(col.name not in self.cols_by_name for col in cols if col.name is not None)
        for col in cols:
            col.tbl = self
            col.id = self.next_col_id
            self.next_col_id += 1

        # we're creating a new schema version
        self.version += 1
        self.created_at = time.time()
        self.schema_version = self.version
        index_cols: dict[Column, tuple[index.BtreeIndex, Column, Column]] = {}
        all_cols: list[Column] = []
        for col in cols:
            all_cols.append(col)
            if col.name is not None and self._is_btree_indexable(col):
                idx = index.BtreeIndex(col)
                val_col, undo_col = self._create_index_columns(idx)
                index_cols[col] = (idx, val_col, undo_col)
                all_cols.append(val_col)
                all_cols.append(undo_col)
        # Add all columns
        status = self._add_columns(all_cols, print_stats=print_stats, on_error=on_error)
        # Create indices and their md records
        for col, (idx, val_col, undo_col) in index_cols.items():
            self._create_index(col, val_col, undo_col, idx_name=None, idx=idx)
        self.update_status = status
        self._write_md(new_version=True, new_schema_version=True)
        _logger.info(f'Added columns {[col.name for col in cols]} to table {self.name}, new version: {self.version}')

        msg = (
            f'Added {status.num_rows} column value{"" if status.num_rows == 1 else "s"} '
            f'with {status.num_excs} error{"" if status.num_excs == 1 else "s"}.'
        )
        Env.get().console_logger.info(msg)
        _logger.info(f'Columns {[col.name for col in cols]}: {msg}')
        return status

    def _add_columns(
        self, cols: Iterable[Column], print_stats: bool, on_error: Literal['abort', 'ignore']
    ) -> UpdateStatus:
        """Add and populate columns within the current transaction"""
        cols_to_add = list(cols)
        row_count = self.store_tbl.count()
        for col in cols_to_add:
            if not col.col_type.nullable and not col.is_computed and row_count > 0:
                raise excs.Error(
                    f'Cannot add non-nullable column {col.name!r} to table {self.name!r} with existing rows'
                )

        computed_values = 0
        num_excs = 0
        cols_with_excs: list[Column] = []
        for col in cols_to_add:
            excs_per_col = 0
            col.schema_version_add = self.schema_version
            # add the column to the lookup structures now, rather than after the store changes executed successfully,
            # because it might be referenced by the next column's value_expr
            self.cols.append(col)
            if col.name is not None:
                self.cols_by_name[col.name] = col
            self.cols_by_id[col.id] = col

            # also add to stored md
            self._tbl_md.column_md[col.id] = schema.ColumnMd(
                id=col.id,
                col_type=col.col_type.as_dict(),
                is_pk=col.is_pk,
                schema_version_add=col.schema_version_add,
                schema_version_drop=col.schema_version_drop,
                value_expr=col.value_expr.as_dict() if col.value_expr is not None else None,
                stored=col.stored,
            )
            if col.name is not None:
                self._schema_version_md.columns[col.id] = schema.SchemaColumn(
                    name=col.name,
                    pos=len(self.cols_by_name),
                    media_validation=col._media_validation.name.lower() if col._media_validation is not None else None,
                )

            if col.is_stored:
                self.store_tbl.add_column(col)

            if not col.is_computed or not col.is_stored or row_count == 0:
                continue

            # populate the column
            from pixeltable.plan import Planner

            plan = Planner.create_add_column_plan(self.path, col)
            plan.ctx.num_rows = row_count
            try:
                plan.open()
                try:
                    excs_per_col = self.store_tbl.load_column(col, plan, on_error == 'abort')
                except sql.exc.DBAPIError as exc:
                    # Wrap the DBAPIError in an excs.Error to unify processing in the subsequent except block
                    raise excs.Error(f'SQL error during execution of computed column `{col.name}`:\n{exc}') from exc
                if excs_per_col > 0:
                    cols_with_excs.append(col)
                    num_excs += excs_per_col
                computed_values += plan.ctx.num_computed_exprs * row_count
            finally:
                # Ensure cleanup occurs if an exception or keyboard interruption happens during `load_column()`.
                def cleanup_on_error() -> None:
                    """Delete columns that are added as part of current add_columns operation and re-initialize
                    the sqlalchemy schema"""
                    self.cols = [col for col in self.cols if col not in cols_to_add]
                    for col in cols_to_add:
                        # remove columns that we already added
                        if col.id in self.cols_by_id:
                            del self.cols_by_id[col.id]
                        if col.name is not None and col.name in self.cols_by_name:
                            del self.cols_by_name[col.name]
                    self.store_tbl.create_sa_tbl()

                # Run cleanup only if there has been an exception; otherwise, skip cleanup.
                run_cleanup_on_exception(cleanup_on_error)
                plan.close()

        pxt.catalog.Catalog.get().record_column_dependencies(self)

        if print_stats:
            plan.ctx.profile.print(num_rows=row_count)

        # TODO: what to do about system columns with exceptions?
        row_counts = RowCountStats(
            upd_rows=row_count, num_excs=num_excs, computed_values=computed_values
        )  # add_columns
        return UpdateStatus(
            cols_with_excs=[f'{col.tbl.name}.{col.name}' for col in cols_with_excs if col.name is not None],
            row_count_stats=row_counts,
        )

    def drop_column(self, col: Column) -> None:
        """Drop a column from the table."""

        assert self.is_mutable

        # we're creating a new schema version
        self.version += 1
        self.created_at = time.time()
        self.schema_version = self.version

        # drop this column and all dependent index columns and indices
        dropped_cols = [col]
        dropped_idx_names: list[str] = []
        for idx_info in self.idxs_by_name.values():
            if idx_info.col != col:
                continue
            dropped_cols.extend([idx_info.val_col, idx_info.undo_col])
            idx_md = self._tbl_md.index_md[idx_info.id]
            idx_md.schema_version_drop = self.schema_version
            assert idx_md.name in self.idxs_by_name
            dropped_idx_names.append(idx_md.name)

        # update idxs_by_name
        for idx_name in dropped_idx_names:
            del self.idxs_by_name[idx_name]

        self._drop_columns(dropped_cols)
        self._write_md(new_version=True, new_schema_version=True)
        _logger.info(f'Dropped column {col.name} from table {self.name}, new version: {self.version}')

    def _drop_columns(self, cols: Iterable[Column]) -> None:
        """Mark columns as dropped"""
        assert self.is_mutable

        for col in cols:
            col.schema_version_drop = self.schema_version
            if col.name is not None:
                assert col.name in self.cols_by_name
                del self.cols_by_name[col.name]
            assert col.id in self.cols_by_id
            del self.cols_by_id[col.id]
            # update stored md
            self._tbl_md.column_md[col.id].schema_version_drop = col.schema_version_drop
            if col.name is not None:
                del self._schema_version_md.columns[col.id]

        # update positions
        for pos, schema_col in enumerate(self._schema_version_md.columns.values()):
            schema_col.pos = pos

        self.store_tbl.create_sa_tbl()
        pxt.catalog.Catalog.get().record_column_dependencies(self)

    def rename_column(self, old_name: str, new_name: str) -> None:
        """Rename a column."""
        assert self.is_mutable
        if old_name not in self.cols_by_name:
            raise excs.Error(f'Unknown column: {old_name}')
        if not is_valid_identifier(new_name):
            raise excs.Error(f"Invalid column name: '{new_name}'")
        if new_name in self.cols_by_name:
            raise excs.Error(f'Column {new_name} already exists')
        col = self.cols_by_name[old_name]
        del self.cols_by_name[old_name]
        col.name = new_name
        self.cols_by_name[new_name] = col
        self._schema_version_md.columns[col.id].name = new_name

        # we're creating a new schema version
        self.version += 1
        self.created_at = time.time()
        self.schema_version = self.version

        self._write_md(new_version=True, new_schema_version=True)
        _logger.info(f'Renamed column {old_name} to {new_name} in table {self.name}, new version: {self.version}')

    def set_comment(self, new_comment: Optional[str]) -> None:
        _logger.info(f'[{self.name}] Updating comment: {new_comment}')
        self.comment = new_comment
        self._create_schema_version()

    def set_num_retained_versions(self, new_num_retained_versions: int) -> None:
        _logger.info(
            f'[{self.name}] Updating num_retained_versions: {new_num_retained_versions} '
            f'(was {self.num_retained_versions})'
        )
        self.num_retained_versions = new_num_retained_versions
        self._create_schema_version()

    def _create_schema_version(self) -> None:
        # we're creating a new schema version
        self.version += 1
        self.created_at = time.time()
        self.schema_version = self.version
        self._write_md(new_version=True, new_schema_version=True)
        _logger.info(f'[{self.name}] Updating table schema to version: {self.version}')

    def insert(
        self,
        rows: Optional[list[dict[str, Any]]],
        df: Optional[pxt.DataFrame],
        print_stats: bool = False,
        fail_on_exception: bool = True,
    ) -> UpdateStatus:
        """
        Insert rows into this table, either from an explicit list of dicts or from a `DataFrame`.
        """
        from pixeltable.plan import Planner

        assert self.is_insertable
        assert (rows is None) != (df is None)  # Exactly one must be specified
        if rows is not None:
            plan = Planner.create_insert_plan(self, rows, ignore_errors=not fail_on_exception)

        else:
            plan = Planner.create_df_insert_plan(self, df, ignore_errors=not fail_on_exception)

        # this is a base table; we generate rowids during the insert
        def rowids() -> Iterator[int]:
            while True:
                rowid = self.next_row_id
                self.next_row_id += 1
                yield rowid

        result = self._insert(
            plan, time.time(), print_stats=print_stats, rowids=rowids(), abort_on_exc=fail_on_exception
        )
        return result

    def _insert(
        self,
        exec_plan: 'exec.ExecNode',
        timestamp: float,
        *,
        rowids: Optional[Iterator[int]] = None,
        print_stats: bool = False,
        abort_on_exc: bool = False,
    ) -> UpdateStatus:
        """Insert rows produced by exec_plan and propagate to views"""
        # we're creating a new version
        self.version += 1
        self.created_at = timestamp
        cols_with_excs, row_counts = self.store_tbl.insert_rows(
            exec_plan, v_min=self.version, rowids=rowids, abort_on_exc=abort_on_exc
        )
        result = UpdateStatus(
            cols_with_excs=[f'{self.name}.{self.cols_by_id[cid].name}' for cid in cols_with_excs],
            row_count_stats=row_counts,
        )

        # update views
        for view in self.mutable_views:
            from pixeltable.plan import Planner

            plan2, _ = Planner.create_view_load_plan(view.get().path, propagates_insert=True)
            status = view.get()._insert(plan2, timestamp, print_stats=print_stats)
            result += status.to_cascade()

        # Use the net status after all propagations
        self.update_status = result
        self._write_md(new_version=True, new_schema_version=False)
        if print_stats:
            exec_plan.ctx.profile.print(num_rows=result.num_rows)
        _logger.info(f'TableVersion {self.name}: new version {self.version}')
        return result

    def update(
        self, value_spec: dict[str, Any], where: Optional[exprs.Expr] = None, cascade: bool = True
    ) -> UpdateStatus:
        """Update rows in this TableVersionPath.
        Args:
            value_spec: a list of (column, value) pairs specifying the columns to update and their new values.
            where: a predicate to filter rows to update.
            cascade: if True, also update all computed columns that transitively depend on the updated columns,
                including within views.
        """
        assert self.is_mutable

        from pixeltable.plan import Planner

        update_spec = self._validate_update_spec(value_spec, allow_pk=False, allow_exprs=True, allow_media=True)
        if where is not None:
            if not isinstance(where, exprs.Expr):
                raise excs.Error(f"'where' argument must be a predicate, got {type(where)}")
            analysis_info = Planner.analyze(self.path, where)
            # for now we require that the updated rows can be identified via SQL, rather than via a Python filter
            if analysis_info.filter is not None:
                raise excs.Error(f'Filter {analysis_info.filter} not expressible in SQL')

        plan, updated_cols, recomputed_cols = Planner.create_update_plan(self.path, update_spec, [], where, cascade)
        from pixeltable.exprs import SqlElementCache

        result = self.propagate_update(
            plan,
            where.sql_expr(SqlElementCache()) if where is not None else None,
            recomputed_cols,
            base_versions=[],
            timestamp=time.time(),
            cascade=cascade,
            show_progress=True,
        )
        result += UpdateStatus(updated_cols=updated_cols)
        return result

    def batch_update(
        self,
        batch: list[dict[Column, exprs.Expr]],
        rowids: list[tuple[int, ...]],
        insert_if_not_exists: bool,
        error_if_not_exists: bool,
        cascade: bool = True,
    ) -> UpdateStatus:
        """Update rows in batch.
        Args:
            batch: one dict per row, each mapping Columns to LiteralExprs representing the new values
            rowids: if not empty, one tuple per row, each containing the rowid values for the corresponding row in batch
        """
        # if we do lookups of rowids, we must have one for each row in the batch
        assert len(rowids) == 0 or len(rowids) == len(batch)

        from pixeltable.plan import Planner

        plan, row_update_node, delete_where_clause, updated_cols, recomputed_cols = Planner.create_batch_update_plan(
            self.path, batch, rowids, cascade=cascade
        )
        result = self.propagate_update(
            plan, delete_where_clause, recomputed_cols, base_versions=[], timestamp=time.time(), cascade=cascade
        )
        result += UpdateStatus(updated_cols=[c.qualified_name for c in updated_cols])

        unmatched_rows = row_update_node.unmatched_rows()
        if len(unmatched_rows) > 0:
            if error_if_not_exists:
                raise excs.Error(f'batch_update(): {len(unmatched_rows)} row(s) not found')
            if insert_if_not_exists:
                insert_status = self.insert(unmatched_rows, None, print_stats=False, fail_on_exception=False)
                result += insert_status.to_cascade()
        return result

    def _validate_update_spec(
        self, value_spec: dict[str, Any], allow_pk: bool, allow_exprs: bool, allow_media: bool
    ) -> dict[Column, exprs.Expr]:
        update_targets: dict[Column, exprs.Expr] = {}
        for col_name, val in value_spec.items():
            if not isinstance(col_name, str):
                raise excs.Error(f'Update specification: dict key must be column name, got {col_name!r}')
            if col_name == _ROWID_COLUMN_NAME:
                # a valid rowid is a list of ints, one per rowid column
                assert len(val) == len(self.store_tbl.rowid_columns())
                for el in val:
                    assert isinstance(el, int)
                continue
            col = self.path.get_column(col_name, include_bases=False)
            if col is None:
                # TODO: return more informative error if this is trying to update a base column
                raise excs.Error(f'Column {col_name} unknown')
            if col.is_computed:
                raise excs.Error(f'Column {col_name} is computed and cannot be updated')
            if col.is_pk and not allow_pk:
                raise excs.Error(f'Column {col_name} is a primary key column and cannot be updated')
            if col.col_type.is_media_type() and not allow_media:
                raise excs.Error(f'Column {col_name} is a media column and cannot be updated')

            # make sure that the value is compatible with the column type
            value_expr: exprs.Expr
            try:
                # check if this is a literal
                value_expr = exprs.Literal(val, col_type=col.col_type)
            except (TypeError, jsonschema.exceptions.ValidationError) as exc:
                if not allow_exprs:
                    raise excs.Error(
                        f'Column {col_name}: value {val!r} is not a valid literal for this column '
                        f'(expected {col.col_type})'
                    ) from exc
                # it's not a literal, let's try to create an expr from it
                value_expr = exprs.Expr.from_object(val)
                if value_expr is None:
                    raise excs.Error(
                        f'Column {col_name}: value {val!r} is not a recognized literal or expression'
                    ) from exc
                if not col.col_type.is_supertype_of(value_expr.col_type, ignore_nullable=True):
                    raise excs.Error(
                        f'Type of value {val!r} ({value_expr.col_type}) is not compatible with the type of column '
                        f'{col_name} ({col.col_type})'
                    ) from exc
            update_targets[col] = value_expr

        return update_targets

    def recompute_columns(self, col_names: list[str], errors_only: bool = False, cascade: bool = True) -> UpdateStatus:
        assert self.is_mutable
        assert all(name in self.cols_by_name for name in col_names)
        assert len(col_names) > 0
        assert len(col_names) == 1 or not errors_only

        from pixeltable.plan import Planner

        target_columns = [self.cols_by_name[name] for name in col_names]
        where_clause: Optional[exprs.Expr] = None
        if errors_only:
            where_clause = (
                exprs.ColumnPropertyRef(exprs.ColumnRef(target_columns[0]), exprs.ColumnPropertyRef.Property.ERRORTYPE)
                != None
            )
        plan, updated_cols, recomputed_cols = Planner.create_update_plan(
            self.path, update_targets={}, recompute_targets=target_columns, where_clause=where_clause, cascade=cascade
        )
        from pixeltable.exprs import SqlElementCache

        result = self.propagate_update(
            plan,
            where_clause.sql_expr(SqlElementCache()) if where_clause is not None else None,
            recomputed_cols,
            base_versions=[],
            timestamp=time.time(),
            cascade=cascade,
            show_progress=True,
        )
        result += UpdateStatus(updated_cols=updated_cols)
        return result

    def propagate_update(
        self,
        plan: Optional[exec.ExecNode],
        where_clause: Optional[sql.ColumnElement],
        recomputed_view_cols: list[Column],
        base_versions: list[Optional[int]],
        timestamp: float,
        cascade: bool,
        show_progress: bool = True,
    ) -> UpdateStatus:
        result = UpdateStatus()
        create_new_table_version = plan is not None
        if create_new_table_version:
            self.version += 1
            self.created_at = timestamp
            cols_with_excs, row_counts = self.store_tbl.insert_rows(
                plan, v_min=self.version, show_progress=show_progress
            )
            result += UpdateStatus(
                row_count_stats=row_counts.insert_to_update(),
                cols_with_excs=[f'{self.name}.{self.cols_by_id[cid].name}' for cid in cols_with_excs],
            )
            self.store_tbl.delete_rows(
                self.version, base_versions=base_versions, match_on_vmin=True, where_clause=where_clause
            )

        if cascade:
            base_versions = [None if plan is None else self.version, *base_versions]  # don't update in place
            # propagate to views
            for view in self.mutable_views:
                recomputed_cols = [col for col in recomputed_view_cols if col.tbl.id == view.id]
                plan = None
                if len(recomputed_cols) > 0:
                    from pixeltable.plan import Planner

                    plan = Planner.create_view_update_plan(view.get().path, recompute_targets=recomputed_cols)
                status = view.get().propagate_update(
                    plan, None, recomputed_view_cols, base_versions=base_versions, timestamp=timestamp, cascade=True
                )
                result += status.to_cascade()
        if create_new_table_version:
            self.update_status = result
            self._write_md(new_version=True, new_schema_version=False)
        return result

    def delete(self, where: Optional[exprs.Expr] = None) -> UpdateStatus:
        """Delete rows in this table.
        Args:
            where: a predicate to filter rows to delete.

        Returns:
            UpdateStatus: an object containing the number of deleted rows and other statistics.
        """
        assert self.is_insertable
        from pixeltable.exprs import Expr
        from pixeltable.plan import Planner

        sql_where_clause: Optional[Expr] = None
        if where is not None:
            if not isinstance(where, Expr):
                raise excs.Error(f"'where' argument must be a predicate, got {type(where)}")
            analysis_info = Planner.analyze(self.path, where)
            # for now we require that the updated rows can be identified via SQL, rather than via a Python filter
            if analysis_info.filter is not None:
                raise excs.Error(f'Filter {analysis_info.filter} not expressible in SQL')
            sql_where_clause = analysis_info.sql_where_clause

        status = self.propagate_delete(sql_where_clause, base_versions=[], timestamp=time.time())
        return status

    def propagate_delete(
        self, where: Optional[exprs.Expr], base_versions: list[Optional[int]], timestamp: float
    ) -> UpdateStatus:
        """Delete rows in this table and propagate to views.
        Args:
            where: a predicate to filter rows to delete.
        Returns:
            number of deleted rows
        """
        # print(f'calling sql_expr()')
        sql_where_clause = where.sql_expr(exprs.SqlElementCache()) if where is not None else None
        # #print(f'sql_where_clause={str(sql_where_clause) if sql_where_clause is not None else None}')
        # sql_cols: list[sql.Column] = []
        # def collect_cols(col) -> None:
        #     sql_cols.append(col)
        # sql.sql.visitors.traverse(sql_where_clause, {}, {'column': collect_cols})
        # x = [f'{str(c)}:{hash(c)}:{id(c.table)}' for c in sql_cols]
        # print(f'where_clause cols: {x}')
        del_rows = self.store_tbl.delete_rows(
            self.version + 1, base_versions=base_versions, match_on_vmin=False, where_clause=sql_where_clause
        )
        row_counts = RowCountStats(del_rows=del_rows)  # delete
        result = UpdateStatus(row_count_stats=row_counts)
        if del_rows > 0:
            # we're creating a new version
            self.version += 1
            self.created_at = timestamp
        for view in self.mutable_views:
            status = view.get().propagate_delete(
                where=None, base_versions=[self.version, *base_versions], timestamp=timestamp
            )
            result += status.to_cascade()
        self.update_status = result

        if del_rows > 0:
            self._write_md(new_version=True, new_schema_version=False)
        return result

    def revert(self) -> None:
        """Reverts the table to the previous version."""
        assert self.is_mutable
        if self.version == 0:
            raise excs.Error('Cannot revert version 0')
        self._revert()

    def _revert(self) -> None:
        """
        Reverts the stored metadata for this table version and propagates to views.

        Doesn't attempt to revert the in-memory metadata, but instead invalidates this TableVersion instance
        and relies on Catalog to reload it
        """
        conn = Env.get().conn
        # make sure we don't have a snapshot referencing this version
        # (unclear how to express this with sqlalchemy)
        query = (
            f"select ts.dir_id, ts.md->'name' "
            f'from {schema.Table.__tablename__} ts '
            f"cross join lateral jsonb_path_query(md, '$.view_md.base_versions[*]') as tbl_version "
            f"where tbl_version->>0 = '{self.id.hex}' and (tbl_version->>1)::int = {self.version}"
        )
        result = list(conn.execute(sql.text(query)))
        if len(result) > 0:
            names = [row[1] for row in result]
            raise excs.Error(
                (
                    f'Current version is needed for {len(result)} snapshot{"s" if len(result) > 1 else ""} '
                    f'({", ".join(names)})'
                )
            )

        # delete newly-added data
        MediaStore.delete(self.id, version=self.version)
        conn.execute(sql.delete(self.store_tbl.sa_tbl).where(self.store_tbl.sa_tbl.c.v_min == self.version))

        # revert new deletions
        set_clause: dict[sql.Column, Any] = {self.store_tbl.sa_tbl.c.v_max: schema.Table.MAX_VERSION}
        for index_info in self.idxs_by_name.values():
            # copy the index value back from the undo column and reset the undo column to NULL
            set_clause[index_info.val_col.sa_col] = index_info.undo_col.sa_col
            set_clause[index_info.undo_col.sa_col] = None
        stmt = sql.update(self.store_tbl.sa_tbl).values(set_clause).where(self.store_tbl.sa_tbl.c.v_max == self.version)
        conn.execute(stmt)

        # revert schema changes:
        # - undo changes to self._tbl_md and write that back
        # - delete newly-added TableVersion/TableSchemaVersion records
        if self.version == self.schema_version:
            # physically delete newly-added columns and remove them from the stored md
            added_cols = [col for col in self.cols if col.schema_version_add == self.schema_version]
            if len(added_cols) > 0:
                self._tbl_md.next_col_id = min(col.id for col in added_cols)
                for col in added_cols:
                    if col.is_stored:
                        self.store_tbl.drop_column(col)
                    del self._tbl_md.column_md[col.id]

            # remove newly-added indices from the lookup structures
            # (the value and undo columns got removed in the preceding step)
            added_idx_md = [md for md in self._tbl_md.index_md.values() if md.schema_version_add == self.schema_version]
            if len(added_idx_md) > 0:
                self._tbl_md.next_idx_id = min(md.id for md in added_idx_md)
                for md in added_idx_md:
                    # TODO: drop the index
                    del self._tbl_md.index_md[md.id]

            # make newly-dropped columns visible again
            dropped_col_md = [
                md for md in self._tbl_md.column_md.values() if md.schema_version_drop == self.schema_version
            ]
            for col_md in dropped_col_md:
                col_md.schema_version_drop = None

            # make newly-dropped indices visible again
            dropped_idx_md = [
                md for md in self._tbl_md.index_md.values() if md.schema_version_drop == self.schema_version
            ]
            for idx_md in dropped_idx_md:
                idx_md.schema_version_drop = None

            conn.execute(
                sql.delete(schema.TableSchemaVersion.__table__)
                .where(schema.TableSchemaVersion.tbl_id == self.id)
                .where(schema.TableSchemaVersion.schema_version == self.schema_version)
            )
            self._tbl_md.current_schema_version = self._schema_version_md.preceding_schema_version

        conn.execute(
            sql.delete(schema.TableVersion.__table__)
            .where(schema.TableVersion.tbl_id == self.id)
            .where(schema.TableVersion.version == self.version)
        )

        self.version -= 1
        self._write_md(new_version=False, new_schema_version=False)

        # propagate to views
        views_str = ', '.join([str(v.id) for v in self.mutable_views])
        print(f'revert(): mutable_views={views_str}')
        for view in self.mutable_views:
            view.get()._revert()

        # force reload on next operation
        self.is_validated = False
        pxt.catalog.Catalog.get().remove_tbl_version(self)
        _logger.info(f'TableVersion {self.name}: reverted to version {self.version}')

    def _init_external_stores(self) -> None:
        for store_md in self.tbl_md.external_stores:
            store_cls = resolve_symbol(store_md['class'])
            assert isinstance(store_cls, type) and issubclass(store_cls, pxt.io.ExternalStore)
            store = store_cls.from_dict(store_md['md'])
            self.external_stores[store.name] = store

    def link_external_store(self, store: pxt.io.ExternalStore) -> None:
        self.version += 1
        self.created_at = time.time()
        self.schema_version = self.version

        self.external_stores[store.name] = store
        self._tbl_md.external_stores.append(
            {'class': f'{type(store).__module__}.{type(store).__qualname__}', 'md': store.as_dict()}
        )
        self._write_md(new_version=True, new_schema_version=True)

    def unlink_external_store(self, store: pxt.io.ExternalStore) -> None:
        del self.external_stores[store.name]
        self.version += 1
        self.created_at = time.time()
        self.schema_version = self.version
        idx = next(i for i, store_md in enumerate(self._tbl_md.external_stores) if store_md['md']['name'] == store.name)
        self._tbl_md.external_stores.pop(idx)
        self._write_md(new_version=True, new_schema_version=True)

    @property
    def tbl_md(self) -> schema.TableMd:
        return self._tbl_md

    @property
    def version_md(self) -> schema.TableVersionMd:
        return self._version_md

    @property
    def schema_version_md(self) -> schema.TableSchemaVersionMd:
        return self._schema_version_md

    @property
    def view_md(self) -> Optional[schema.ViewMd]:
        return self._tbl_md.view_md

    @property
    def name(self) -> str:
        return self._tbl_md.name

    @property
    def user(self) -> Optional[str]:
        return self._tbl_md.user

    @property
    def is_replica(self) -> bool:
        return self._tbl_md.is_replica

    @property
    def comment(self) -> str:
        return self._schema_version_md.comment

    @comment.setter
    def comment(self, c: str) -> None:
        assert self.effective_version is None
        self._schema_version_md.comment = c

    @property
    def num_retained_versions(self) -> int:
        return self._schema_version_md.num_retained_versions

    @num_retained_versions.setter
    def num_retained_versions(self, n: int) -> None:
        assert self.effective_version is None
        self._schema_version_md.num_retained_versions = n

    @property
    def version(self) -> int:
        # if this is a snapshot instance, we need to ignore current_version
        return self._tbl_md.current_version if self.effective_version is None else self.effective_version

    @version.setter
    def version(self, version: int) -> None:
        assert self.effective_version is None
        self._tbl_md.current_version = version
        self._version_md.version = version

    @property
    def created_at(self) -> float:
        return self._version_md.created_at

    @created_at.setter
    def created_at(self, ts: float) -> None:
        assert self.effective_version is None
        self._version_md.created_at = ts

    @property
    def schema_version(self) -> int:
        return self._schema_version_md.schema_version

    @schema_version.setter
    def schema_version(self, version: int) -> None:
        assert self.effective_version is None
        self._tbl_md.current_schema_version = version
        self._version_md.schema_version = version
        self._schema_version_md.preceding_schema_version = self._schema_version_md.schema_version
        self._schema_version_md.schema_version = version

    @property
    def preceding_schema_version(self) -> int:
        return self._schema_version_md.preceding_schema_version

    @property
    def update_status(self) -> Optional[UpdateStatus]:
        return self._version_md.update_status

    @update_status.setter
    def update_status(self, status: UpdateStatus) -> None:
        assert self.effective_version is None
        self._version_md.update_status = status

    @property
    def media_validation(self) -> MediaValidation:
        return MediaValidation[self._schema_version_md.media_validation.upper()]

    @property
    def next_col_id(self) -> int:
        return self._tbl_md.next_col_id

    @next_col_id.setter
    def next_col_id(self, id: int) -> None:
        assert self.effective_version is None
        self._tbl_md.next_col_id = id

    @property
    def next_idx_id(self) -> int:
        return self._tbl_md.next_idx_id

    @next_idx_id.setter
    def next_idx_id(self, id: int) -> None:
        assert self.effective_version is None
        self._tbl_md.next_idx_id = id

    @property
    def next_row_id(self) -> int:
        return self._tbl_md.next_row_id

    @next_row_id.setter
    def next_row_id(self, id: int) -> None:
        assert self.effective_version is None
        self._tbl_md.next_row_id = id

    @property
    def is_snapshot(self) -> bool:
        return self.effective_version is not None

    @property
    def is_mutable(self) -> bool:
        return not self.is_snapshot and not self.is_replica

    @property
    def is_view(self) -> bool:
        return self.view_md is not None

    @property
    def include_base_columns(self) -> bool:
        return self.view_md is not None and self.view_md.include_base_columns

    @property
    def is_component_view(self) -> bool:
        return self.iterator_cls is not None

    @property
    def is_insertable(self) -> bool:
        """Returns True if this corresponds to an InsertableTable"""
        return self.is_mutable and not self.is_view

    def is_iterator_column(self, col: Column) -> bool:
        """Returns True if col is produced by an iterator"""
        # the iterator columns directly follow the pos column
        return self.is_component_view and col.id > 0 and col.id < self.num_iterator_cols + 1

    def is_system_column(self, col: Column) -> bool:
        """Return True if column was created by Pixeltable"""
        return col.name == _POS_COLUMN_NAME and self.is_component_view

    def user_columns(self) -> list[Column]:
        """Return all non-system columns"""
        return [c for c in self.cols if not self.is_system_column(c)]

    def primary_key_columns(self) -> list[Column]:
        """Return all non-system columns"""
        return [c for c in self.cols if c.is_pk]

    @property
    def primary_key(self) -> list[str]:
        """Return the names of the primary key columns"""
        return [c.name for c in self.cols if c.is_pk]

    def get_required_col_names(self) -> list[str]:
        """Return the names of all columns for which values must be specified in insert()"""
        assert not self.is_view
        names = [c.name for c in self.cols_by_name.values() if not c.is_computed and not c.col_type.nullable]
        return names

    def get_computed_col_names(self) -> list[str]:
        """Return the names of all computed columns"""
        names = [c.name for c in self.cols_by_name.values() if c.is_computed]
        return names

    def get_idx_val_columns(self, cols: Iterable[Column]) -> set[Column]:
        result = {info.val_col for col in cols for info in col.get_idx_info().values()}
        return result

    def get_dependent_columns(self, cols: Iterable[Column]) -> set[Column]:
        """
        Return the set of columns that transitively depend on any of the given ones.
        """
        cat = pxt.catalog.Catalog.get()
        result = set().union(*[cat.get_column_dependents(col.tbl.id, col.id) for col in cols])
        if len(result) > 0:
            result.update(self.get_dependent_columns(result))
        return result

    def num_rowid_columns(self) -> int:
        """Return the number of columns of the rowids, without accessing store_tbl"""
        if self.is_component_view:
            return 1 + self.base.get().num_rowid_columns()
        return 1

    @classmethod
    def _create_column_md(cls, cols: list[Column]) -> dict[int, schema.ColumnMd]:
        column_md: dict[int, schema.ColumnMd] = {}
        for col in cols:
            value_expr_dict = col.value_expr.as_dict() if col.value_expr is not None else None
            assert col.is_pk is not None
            column_md[col.id] = schema.ColumnMd(
                id=col.id,
                col_type=col.col_type.as_dict(),
                is_pk=col.is_pk,
                schema_version_add=col.schema_version_add,
                schema_version_drop=col.schema_version_drop,
                value_expr=value_expr_dict,
                stored=col.stored,
            )
        return column_md

    @classmethod
    def _create_stores_md(cls, stores: Iterable[pxt.io.ExternalStore]) -> list[dict[str, Any]]:
        return [
            {'class': f'{type(store).__module__}.{type(store).__qualname__}', 'md': store.as_dict()} for store in stores
        ]

    def as_dict(self) -> dict:
        return {'id': str(self.id), 'effective_version': self.effective_version}

    @classmethod
    def from_dict(cls, d: dict) -> TableVersion:
        from pixeltable import catalog

        id = UUID(d['id'])
        effective_version = d['effective_version']
        return catalog.Catalog.get().get_tbl_version(id, effective_version)
